# -*- coding: utf-8 -*-
"""Client for Jira

.. module:: trackapps.jira
   :platform: Unix
   :synopsis: Client for Jira
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
track_before_connect
track_after_connect
track_before_read
track_after_read
track_before_create
track_after_create
track_before_update
track_after_update

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from hydratk.lib.network.rest.client import RESTClient
from simplejson import dumps

config = {
    'session': '/rest/auth/1/session',
    'search': '/rest/api/2/search',
    'issue': '/rest/api/2/issue'
}


class Client(object):
    """Class Client
    """

    _mh = None
    _client = None
    _url = None
    _user = None
    _passw = None
    _project = None
    _return_fields = None
    _default_values = {}
    _is_connected = None

    def __init__(self):
        """Class constructor

        Called when the object is initialized   

        Args:
           none

        """

        self._mh = MasterHead.get_head()
        self._client = RESTClient()

        cfg = self._mh.cfg['Extensions']['TrackApps']['jira']
        if ('return_fields' in cfg and cfg['return_fields'] != None):
            self._return_fields = cfg['return_fields'].split(',')
        if ('default_values' in cfg and cfg['default_values'] != None):
            self._default_values = cfg['default_values']
        if ('url' in cfg and cfg['url'] != None):
            self._url = cfg['url']
        if ('user' in cfg and cfg['user'] != None):
            self._user = cfg['user']
        if ('passw' in cfg and cfg['passw'] != None):
            self._passw = cfg['passw']
        if ('project' in cfg and cfg['project'] != None):
            self._project = cfg['project']

    @property
    def client(self):
        """ client property getter """

        return self._client

    @property
    def url(self):
        """ url property getter """

        return self._url

    @property
    def user(self):
        """ user property getter """

        return self._user

    @property
    def passw(self):
        """ passw property getter """

        return self._passw

    @property
    def project(self):
        """ project property getter """

        return self._project

    @property
    def return_fields(self):
        """ return_fields property getter """

        return self._return_fields

    @property
    def default_values(self):
        """ default_values property getter """

        return self._default_values

    @property
    def is_connected(self):
        """ is_connected property getter """

        return self._is_connected

    def connect(self, url=None, user=None, passw=None, project=None):
        """Method connects to Jira

        Args:    
           url (str): URL        
           user (str): username
           passw (str): password
           project (str): project

        Returns:
           bool: result

        Raises:
           event: track_before_connect
           event: track_after_connect

        """

        message = 'url:{0}, user:{1}, passw:{2}, project:{3}'.format(
            url, user, passw, project)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_connecting', message), self._mh.fromhere())

        if (url == None):
            url = self._url
        if (user == None):
            user = self._user
        if (passw == None):
            passw = self._passw
        if (project == None):
            project = self._project

        ev = event.Event('track_before_connect', url, user, passw, project)
        if (self._mh.fire_event(ev) > 0):
            url = ev.argv(0)
            user = ev.argv(1)
            passw = ev.argv(2)
            project = ev.argv(3)

        if (ev.will_run_default()):
            self._url = url
            self._user = user
            self._passw = passw
            self._project = project

            url = self._url + config['session']
            body = {'username': self._user, 'password': self._passw}
            res, body = self._client.send_request(url, method='POST', headers={'Accept': 'application/json'},
                                                  body=dumps(body), content_type='json')

        result = False
        if (res == 200):
            self._is_connected = True
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_connected'), self._mh.fromhere())
            ev = event.Event('track_after_connect')
            self._mh.fire_event(ev)
            result = True
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return result

    def read(self, id=None, fields=None, query=None, limit=None, offset=None):
        """Method reads records

        Args: 
           id (int): record id         
           fields (list): fields to be returned, default all
           query (str): record query, see Jira doc
           limit (int): record count    
           offset (int): record offset

        Returns:
           tuple: result (bool), records (list of dictionary)

        Raises:
           event: track_before_read
           event: track_after_read

        """

        message = 'id:{0}, fields:{1}, query:{2}, limit:{3}, offset:{4}'.format(
            id, fields, query, limit, offset)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_reading', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and self._return_fields != None):
            fields = self._return_fields
        ev = event.Event('track_before_read', id, fields, query, limit, offset)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            fields = ev.argv(1)
            query = ev.argv(2)
            limit = ev.argv(3)
            offset = ev.argv(4)

        if (ev.will_run_default()):

            body = {}
            if (id != None):
                body['jql'] = 'key={0}-{1}'.format(self._project, id)
            else:
                body['jql'] = 'project=' + self._project
            if (fields != None and len(fields) > 0):
                fields_new = []
                for field in fields:
                    fields_new.append(field)
                body['fields'] = fields_new
            if (query != None):
                body['jql'] += ' and ' + query
            if (limit != None):
                body['maxResults'] = limit
            if (offset != None):
                body['startAt'] = offset
            body = dumps(body)

            url = self._url + config['search']
            res, body = self._client.send_request(url, method='POST', user=self._user, passw=self._passw,
                                                  headers={'Accept': 'application/json'}, body=body, content_type='json')

            result = False
            records = None
            if (res == 200):
                cnt2 = len(body['issues'])
                cnt = body['total'] if (body['total'] == cnt2) else cnt2
                records = []
                for i in range(0, cnt):
                    record = {}
                    if (fields == None or 'id' in fields):
                        record['id'] = int(
                            body['issues'][i]['key'].split('-')[-1])
                    for key, value in body['issues'][i]['fields'].items():
                        if (fields == None or key in fields):
                            record[key] = value
                    if (record != {}):
                        records.append(record)

                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_read', cnt), self._mh.fromhere())
                ev = event.Event('track_after_read')
                self._mh.fire_event(ev)
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', res, body), self._mh.fromhere())

            return (result, records)

    def create(self, params={}):
        """Method creates record

        Args: 
           params (dict): record content, key - field name, value - field value

        Returns:
           int: record id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'issue', params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        if (self._default_values != {}):
            for key, value in self._default_values.items():
                if (key not in params):
                    params[key] = value

        ev = event.Event('track_before_create', params)
        if (self._mh.fire_event(ev) > 0):
            params = ev.argv(0)

        if (ev.will_run_default()):

            root = {'fields': {}}
            for key, value in params.items():
                root['fields'][key] = value
            if ('project' not in root['fields']):
                root['fields']['project'] = {'key': self._project}
            if ('issuetype' not in root['fields']):
                root['fields']['issuetype'] = {'name': 'Bug'}
            body = dumps(root)

            url = self._url + config['issue']
            res, body = self._client.send_request(url, method='POST', user=self._user, passw=self._passw,
                                                  headers={'Accept': 'application/json'}, body=body, content_type='json')

        id = None
        if (res in (200, 201)):
            id = int(body['key'].split('-')[1])
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_created', id), self._mh.fromhere())
            ev = event.Event('track_after_create')
            self._mh.fire_event(ev)
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return id

    def update(self, id, params={}):
        """Method updates record

        Args: 
           id (int): record id
           params (dict): record content, key - field name, value - field value

        Returns:
           bool: result

        Raises:
           event: track_before_update
           event: track_after_update

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_updating', 'issue', id, params),
                      self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_update', id, params)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            params = ev.argv(1)

        if (ev.will_run_default()):

            root = {'fields': {}}
            for key, value in params.items():
                root['fields'][key] = value
            body = dumps(root)

            url = self._url + config['issue'] + \
                '/{0}-{1}'.format(self._project, id)
            res, body = self._client.send_request(url, method='POST', user=self._user, passw=self._passw,
                                                  headers={'Accept': 'application/json'}, body=body, content_type='json')

        result = False
        if (res in (200, 204)):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_updated', id), self._mh.fromhere())
            ev = event.Event('track_after_update')
            self._mh.fire_event(ev)
            result = True
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return result



# -*- coding: utf-8 -*-
"""Client for Bugzilla

.. module:: trackapps.bugzilla
   :platform: Unix
   :synopsis: Client for Bugzilla
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
track_before_connect
track_after_connect
track_before_read
track_after_read
track_before_create
track_after_create
track_before_update
track_after_update

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from hydratk.lib.network.rest.client import RESTClient
from simplejson import dumps

config = {
    'login': '/rest.cgi/login',
    'logout': '/rest.cgi/logout',
    'rest': '/rest.cgi/bug'
}


class Client(object):
    """Class Client
    """

    _mh = None
    _client = None
    _url = None
    _user = None
    _passw = None
    _token = None
    _return_fields = None
    _default_values = {}
    _is_connected = None

    def __init__(self):
        """Class constructor

        Called when the object is initialized

        Args:   
           none

        """

        self._mh = MasterHead.get_head()
        self._client = RESTClient()

        cfg = self._mh.cfg['Extensions']['TrackApps']['bugzilla']
        if ('return_fields' in cfg and cfg['return_fields'] != None):
            self._return_fields = cfg['return_fields'].split(',')
        if ('default_values' in cfg and cfg['default_values'] != None):
            self._default_values = cfg['default_values']
        if ('url' in cfg and cfg['url'] != None):
            self._url = cfg['url']
        if ('user' in cfg and cfg['user'] != None):
            self._user = cfg['user']
        if ('passw' in cfg and cfg['passw'] != None):
            self._passw = cfg['passw']

    @property
    def client(self):
        """ client property getter """

        return self._client

    @property
    def url(self):
        """ url property getter """

        return self._url

    @property
    def user(self):
        """ user property getter """

        return self._user

    @property
    def passw(self):
        """ passw property getter """

        return self._passw

    @property
    def token(self):
        """ token property getter """

        return self._token

    @property
    def return_fields(self):
        """ return_fields property getter """

        return self._return_fields

    @property
    def default_values(self):
        """ default_values property getter """

        return self._default_values

    @property
    def is_connected(self):
        """ is_conncted property getter """

        return self._is_connected

    def connect(self, url=None, user=None, passw=None):
        """Method connects to Bugzilla

        Args:    
           url (str): URL        
           user (str): username
           passw (str): password

        Returns:
           bool: result

        Raises:
           event: track_before_connect
           event: track_after_connect

        """

        message = 'url:{0}, user:{1}, passw:{2}'.format(url, user, passw)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_connecting', message), self._mh.fromhere())

        if (url == None):
            url = self._url
        if (user == None):
            user = self._user
        if (passw == None):
            passw = self._passw

        ev = event.Event('track_before_connect', url, user, passw)
        if (self._mh.fire_event(ev) > 0):
            url = ev.argv(0)
            user = ev.argv(1)
            passw = ev.argv(2)

        if (ev.will_run_default()):
            self._url = url
            self._user = user
            self._passw = passw

            url = self._url + config['login']
            params = {'login': self._user, 'password': self._passw}
            res, body = self._client.send_request(url, method='GET', headers={'Accept': 'application/json'},
                                                  params=params, content_type='json')

        result = False
        if (res == 200):
            self._token = body['token']
            if (self._token != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_connected'), self._mh.fromhere())
                ev = event.Event('track_after_connect')
                self._mh.fire_event(ev)
                self._is_connected = True
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_missing_token'), self._mh.fromhere())
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return result

    def disconnect(self):
        """Method disconnects from Bugzilla

        Args:  
           none  

        Returns:
           bool: result

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_disconnecting'), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        url = self._url + config['logout']
        res, body = self._client.send_request(
            url, method='GET', params={'token': self._token})

        result = False
        if (res == 200):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_disconnected'), self._mh.fromhere())
            self._token = None
            self._is_connected = False
            result = True
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return result

    def read(self, id=None, fields=None, query=None, limit=None, offset=None):
        """Method reads records

        Args: 
           id (int): record id         
           fields (list): fields to be returned, default all
           query (dict): record query, see Bugzilla doc
           limit (int): record count    
           offset (int): record offset

        Returns:
           tuple: result (bool), records (list of dictionary)

        Raises:
           event: track_before_read
           event: track_after_read

        """

        message = 'id:{0}, fields:{1}, query:{2}, limit:{3}, offset:{4}'.format(
            id, fields, query, limit, offset)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_reading', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and self._return_fields != None):
            fields = self._return_fields

        ev = event.Event('track_before_read', id, fields, query, limit, offset)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            fields = ev.argv(1)
            query = ev.argv(2)
            limit = ev.argv(3)
            offset = ev.argv(4)

        if (ev.will_run_default()):

            params = {}
            if (query != None):
                for key, val in query.items():
                    params[key] = val
            if (id != None):
                params['id'] = id
            if (limit != None):
                params['limit'] = limit
            if (offset != None):
                params['offset'] = offset

            url = self._url + config['rest']
            params['token'] = self._token
            res, body = self._client.send_request(url, method='GET', headers={'Accept': 'application/json'},
                                                  params=params, content_type='json')
            result = False
            records = None
            if (res == 200):

                cnt = len(body['bugs'])
                records = []
                for i in range(0, cnt):
                    record = {}
                    for key, value in body['bugs'][i].items():
                        if (fields == None or key in fields):
                            record[key] = value
                    if (record != {}):
                        records.append(record)

                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_read', cnt), self._mh.fromhere())
                ev = event.Event('track_after_read')
                self._mh.fire_event(ev)
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', res, body), self._mh.fromhere())

            return (result, records)

    def create(self, params={}):
        """Method creates record

        Args: 
           params (dict): record content, key - field name, value - field value

        Returns:
           int: record id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'bug', params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        if (self._default_values != {}):
            for key, value in self._default_values.items():
                if (key not in params):
                    params[key] = value

        ev = event.Event('track_before_create', params)
        if (self._mh.fire_event(ev) > 0):
            params = ev.argv(0)

        if (ev.will_run_default()):

            root = {}
            for key, value in params.items():
                root[key] = value
            root['token'] = self._token
            body = dumps(root)

            url = self._url + config['rest']
            res, body = self._client.send_request(url, method='POST', headers={'Accept': 'application/json'},
                                                  body=body, content_type='json')

        id = None
        if (res == 200):
            id = body['id']
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_created', id), self._mh.fromhere())
            ev = event.Event('track_after_create')
            self._mh.fire_event(ev)
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return id

    def update(self, id, params={}):
        """Method updates record

        Args: 
           id (int): record id          
           params (dict): record content, key - field name, value - field value 

        Returns:
           bool: result

        Raises:
           event: track_before_update
           event: track_after_update

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_updating', 'bug', id, params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_update', id, params)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            params = ev.argv(1)

        if (ev.will_run_default()):

            root = {}
            for key, value in params.items():
                root[key] = value
            root['token'] = self._token
            body = dumps(root)

            url = self._url + config['rest'] + '/' + str(id)
            res, body = self._client.send_request(url, method='PUT', headers={'Accept': 'application/json'},
                                                  body=body, content_type='json')

        result = False
        if (res == 200):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_updated', id), self._mh.fromhere())
            ev = event.Event('track_after_update')
            self._mh.fire_event(ev)
            result = True
        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, body), self._mh.fromhere())

        return result



# -*- coding: utf-8 -*-
"""Extension with interface to bugtracking and test management systems

.. module:: trackapps.trackapps
   :platform: Unix
   :synopsis: Extension with interface to bugtracking and test management systems
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

from hydratk.core import extension, bootstrapper
from hydratk.lib.console.commandlinetool import CommandlineTool
from os import path
from importlib import import_module
import hydratk.lib.system.config as syscfg

dep_modules = {
    'hydratk': {
        'min-version': '0.5.0',
        'package': 'hydratk'
    },
    'hydratk.lib.network': {
        'min-version': '0.2.0',
        'package': 'hydratk-lib-network'
    }
}

apps = {
    'qc': 'qc',
    'bugzilla': 'bugzilla',
    'mantis': 'mantis',
    'trac': 'trac',
    'jira': 'jira',
    'testlink': 'testlink'
}


class Extension(extension.Extension):
    """Class Extension
    """

    def _init_extension(self):
        """Method initializes extension

        Args:
           none

        Returns:
           void

        """

        self._ext_id = 'trackapps'
        self._ext_name = 'TrackApps'
        self._ext_version = '0.1.1'
        self._ext_author = 'Petr Rašek <bowman@hydratk.org>, HydraTK team <team@hydratk.org>'
        self._ext_year = '2016-2018'

        if (not self._check_dependencies()):
            exit(0)

    def _check_dependencies(self):
        """Method checks dependent modules

        Args:            
           none

        Returns:
           bool    

        """

        return bootstrapper._check_dependencies(dep_modules, 'hydratk-ext-trackapps')

    def _uninstall(self):
        """Method returns additional uninstall data 

        Args:            
           none

        Returns:
           tuple: list (files), list (modules)   

        """

        files = [
            '/usr/share/man/man1/trackapps.1',
            '{0}/hydratk/conf.d/hydratk-ext-trackapps.conf'.format(syscfg.HTK_ETC_DIR)
        ]

        return files, dep_modules

    def _register_actions(self):
        """Method registers command hooks

        Args:
           none

        Returns:
           void

        """

        if (self._mh.cli_cmdopt_profile == 'trackapps'):
            self._register_standalone_actions()
        else:
            self._register_htk_actions()

    def _register_htk_actions(self):
        """Method registers command hooks

        Args:  
           none        

        Returns:
           void

        """

        self._mh.match_cli_command('track')

        hook = [
            {'command': 'track', 'callback': self.handle_track}
        ]
        self._mh.register_command_hook(hook)

        self._mh.match_long_option('tr-app', True, 'tr-app')
        self._mh.match_long_option('tr-action', True, 'tr-action')
        self._mh.match_long_option('tr-type', True, 'tr-type')
        self._mh.match_long_option('tr-input', True, 'tr-input')
        self._mh.match_long_option('tr-output', True, 'tr-output')
        self._mh.match_long_option('tr-url', True, 'tr-url')
        self._mh.match_long_option('tr-user', True, 'tr-user')
        self._mh.match_long_option('tr-passw', True, 'tr-passw')
        self._mh.match_long_option('tr-dev-key', True, 'tr-dev-key')
        self._mh.match_long_option('tr-domain', True, 'tr-domain')
        self._mh.match_long_option('tr-project', True, 'tr-project')
        self._mh.match_long_option('tr-id', True, 'tr-id')
        self._mh.match_long_option('tr-fields', True, 'tr-fields')
        self._mh.match_long_option('tr-query', True, 'tr-query')
        self._mh.match_long_option('tr-order-by', True, 'tr-order-by')
        self._mh.match_long_option('tr-limit', True, 'tr-limit')
        self._mh.match_long_option('tr-offset', True, 'tr-offset')
        self._mh.match_long_option('tr-page', True, 'tr-page')
        self._mh.match_long_option('tr-per-page', True, 'tr-per-page')
        self._mh.match_long_option('tr-params', True, 'tr-params')
        self._mh.match_long_option('tr-path', True, 'tr-path')
        self._mh.match_long_option('tr-steps', True, 'tr-steps')

    def _register_standalone_actions(self):
        """Method registers command hooks for standalone mode

        Args:  
           none        

        Returns:
           void

        """

        option_profile = 'trackapps'
        help_title = '{h}' + self._ext_name + ' v' + self._ext_version + '{e}'
        cp_string = '{u}' + "(c) " + self._ext_year + \
            " " + self._ext_author + '{e}'
        self._mh.set_cli_appl_title(help_title, cp_string)

        self._mh.match_cli_command('run', option_profile)

        hook = [
            {'command': 'run', 'callback': self.handle_track}
        ]
        self._mh.register_command_hook(hook)

        self._mh.match_cli_command('help', option_profile)

        self._mh.match_long_option(
            'app', True, 'tr-app', False, option_profile)
        self._mh.match_long_option(
            'action', True, 'tr-action', False, option_profile)
        self._mh.match_long_option(
            'type', True, 'tr-type', False, option_profile)
        self._mh.match_long_option(
            'input', True, 'tr-input', False, option_profile)
        self._mh.match_long_option(
            'output', True, 'tr-output', False, option_profile)
        self._mh.match_long_option(
            'url', True, 'tr-url', False, option_profile)
        self._mh.match_long_option(
            'user', True, 'tr-user', False, option_profile)
        self._mh.match_long_option(
            'passw', True, 'tr-passw', False, option_profile)
        self._mh.match_long_option(
            'dev-key', True, 'tr-dev-key', False, option_profile)
        self._mh.match_long_option(
            'domain', True, 'tr-domain', False, option_profile)
        self._mh.match_long_option(
            'project', True, 'tr-project', False, option_profile)
        self._mh.match_long_option('id', True, 'tr-id', False, option_profile)
        self._mh.match_long_option(
            'fields', True, 'tr-fields', False, option_profile)
        self._mh.match_long_option(
            'query', True, 'tr-query', False, option_profile)
        self._mh.match_long_option(
            'order-by', True, 'tr-order-by', False, option_profile)
        self._mh.match_long_option(
            'limit', True, 'tr-limit', False, option_profile)
        self._mh.match_long_option(
            'offset', True, 'tr-offset', False, option_profile)
        self._mh.match_long_option(
            'page', True, 'tr-page', False, option_profile)
        self._mh.match_long_option(
            'per-page', True, 'tr-per-page', False, option_profile)
        self._mh.match_long_option(
            'params', True, 'tr-params', False, option_profile)
        self._mh.match_long_option(
            'path', True, 'tr-path', False, option_profile)
        self._mh.match_long_option(
            'steps', True, 'tr-steps', False, option_profile)

        self._mh.match_cli_option(
            ('c', 'config'), True, 'config', False, option_profile)
        self._mh.match_cli_option(
            ('d', 'debug'), True, 'debug', False, option_profile)
        self._mh.match_cli_option(
            ('e', 'debug-channel'), True, 'debug-channel', False, option_profile)
        self._mh.match_cli_option(
            ('l', 'language'), True, 'language', False, option_profile)
        self._mh.match_cli_option(
            ('m', 'run-mode'), True, 'run-mode', False, option_profile)
        self._mh.match_cli_option(
            ('f', 'force'), False, 'force', False, option_profile)
        self._mh.match_cli_option(
            ('i', 'interactive'), False, 'interactive', False, option_profile)
        self._mh.match_cli_option(
            ('h', 'home'), False, 'home', False, option_profile)

    def init_client(self, app, *args, **kwargs):
        """Client factory method

        Args:            
            app (str): application, qc|bugzilla|mantis|trac|jira|testlink
            args (args): arguments 
            kwargs (kwargs): key value arguments

        Returns:
            obj: Client

        Raises:
            error: NotImplementedError

        """

        self._app = app.lower()
        if (self._app in apps):
            mod = import_module(
                'hydratk.extensions.trackapps.{0}'.format(apps[app]))
            return mod.Client(*args, **kwargs)
        else:
            raise NotImplementedError('Unknown application:{0}'.format(app))

    def handle_track(self):
        """Method handles command track

        Args:
           none

        Returns:
           void   

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_received_cmd', 'track'), self._mh.fromhere())

        app = CommandlineTool.get_input_option('tr-app')
        action = CommandlineTool.get_input_option('tr-action')

        if (not app):
            print('Missing option app')
        elif (app.lower() not in apps):
            print('Application not in qc|bugzilla|mantis|trac|jira|testlink')
        elif (not action):
            print('Missing option action')
        elif (action not in ['read', 'create', 'update', 'delete']):
            print('Action not in read|create|update|delete')
        elif (action == 'delete' and app.lower() not in ['qc', 'mantis', 'trac']):
            print('Application {0} doesn\'t support {1}'.format(app, action))
        else:

            self._client = self.init_client(app)
            self._cfg = self._mh.cfg['Extensions']['TrackApps'][self._app]

            url = CommandlineTool.get_input_option('tr-url')
            if (not url):
                if ('url' in self._cfg and self._cfg['url'] != None):
                    url = self._cfg['url']
                else:
                    print('Enter url')
                    url = raw_input(':')

            user = CommandlineTool.get_input_option('tr-user')
            if (not user and self._app != 'testlink'):
                if ('user' in self._cfg and self._cfg['user'] != None):
                    user = self._cfg['user']
                else:
                    print('Enter username')
                    user = raw_input(':')

            passw = CommandlineTool.get_input_option('tr-passw')
            if (not passw and self._app != 'testlink'):
                if ('passw' in self._cfg and self._cfg['passw'] != None):
                    passw = self._cfg['passw']
                else:
                    print('Enter password')
                    passw = raw_input(':')

            dev_key = CommandlineTool.get_input_option('tr-dev-key')
            if (not dev_key and self._app == 'testlink'):
                if ('dev_key' in self._cfg and self._cfg['dev_key'] != None):
                    dev_key = self._cfg['dev_key']
                else:
                    print('Enter dev_key')
                    dev_key = raw_input(':')

            domain = CommandlineTool.get_input_option('tr-domain')
            if (not domain and self._app == 'qc'):
                if ('domain' in self._cfg and self._cfg['domain'] != None):
                    domain = self._cfg['domain']
                else:
                    print('Enter domain')
                    domain = raw_input(':')

            project = CommandlineTool.get_input_option('tr-project')
            if (not project and self._app in ['qc', 'mantis', 'trac', 'jira', 'testlink']):
                if ('project' in self._cfg and self._cfg['project'] != None):
                    project = self._cfg['project']
                else:
                    print('Enter project')
                    project = raw_input(':')

            self._entity = CommandlineTool.get_input_option('tr-type')
            if (not self._entity):
                self._entity = 'defect'
            if (self._app == 'qc' and self._entity not in ['defect', 'test-folder', 'test', 'test-set-folder', 'test-set', 'test-instance']):
                print(
                    'Type not in defect|test-folder|test|test-set-folder|test-set|test-instance')
                return
            elif (self._app == 'testlink'):
                if (self._entity not in ['test', 'test-suite', 'test-plan', 'build']):
                    print('Type not in test|test-suite|test-plan|build')
                    return
                elif (action == 'read' and self._entity not in ['test', 'test-suite', 'test-plan']):
                    print(
                        'Type not in test|test-suite|test-plan for action read')
                    return
                elif (action == 'update' and self._entity not in ['test', 'test-plan']):
                    print('Type not in test|test-plan for action update')
                    return

            if (self._app == 'qc'):
                res = self._client.connect(url, user, passw, domain, project)
            elif (self._app == 'bugzilla'):
                res = self._client.connect(url, user, passw)
            elif (self._app in ['mantis', 'trac', 'jira']):
                res = self._client.connect(url, user, passw, project)
            elif (self._app == 'testlink'):
                res = self._client.connect(url, dev_key, project)

            if (not res):
                print('Connection error')
                return

            if (action == 'read'):
                self.read()
            elif (action == 'create'):
                self.create()
            elif (action == 'update'):
                self.update()
            elif (action == 'delete'):
                self.delete()

            if (self._app in ['qc', 'bugzilla']):
                self._client.disconnect()

    def read(self):
        """Method handles read action

        Args:
           none

        Returns:
           void

        """

        id = CommandlineTool.get_input_option('tr-id')
        if (not id):
            id = None

        fields = CommandlineTool.get_input_option('tr-fields')
        if (not fields):
            fields = None

        query = CommandlineTool.get_input_option('tr-query')
        if (not query):
            query = None

        order_by_in = CommandlineTool.get_input_option('tr-order-by')
        if (order_by_in != False):
            order_by = {}
            order_by_in = order_by_in.split(',')
            for param in order_by_in:
                rec = param.split(':')
                key = rec[0]
                order_by[key] = rec[1]
        else:
            order_by = None

        limit = CommandlineTool.get_input_option('tr-limit')
        if (not limit):
            limit = None

        offset = CommandlineTool.get_input_option('tr-offset')
        if (not offset):
            offset = None

        page = CommandlineTool.get_input_option('tr-page')
        if (not page):
            page = -1

        per_page = CommandlineTool.get_input_option('tr-per-page')
        if (not per_page):
            per_page = -1

        if (self._app == 'qc'):
            if (self._entity in ['test-folder', 'test-set-folder']):
                path = CommandlineTool.get_input_option('tr-path')
                if (not path):
                    print('Enter path')
                    path = raw_input(':')
                res, records = self._client.read_test_folder(
                    path, self._entity)
            elif (self._entity == 'test-set'):
                if (not id):
                    print('Enter test set id')
                    id = raw_input(':')
                res, records = self._client.read_test_set(int(id))
            else:
                res, records = self._client.read(
                    id, self._entity, fields, query, order_by, limit, offset)

        elif (self._app == 'testlink'):
            if (self._entity == 'test-suite'):
                path = CommandlineTool.get_input_option('tr-path')
                if (not path):
                    print('Enter path')
                    path = raw_input(':')
                res, records = self._client.read_test_suite(
                    path, fields=fields)
            elif (self._entity == 'test-plan'):
                if (not id):
                    print('Enter plan id')
                    id = raw_input(':')
                res, records = self._client.read_test_plan(
                    plan_id=id, fields=fields)
            elif (self._entity == 'test'):
                if (not id):
                    print('Enter test id')
                    id = raw_input(':')
                res, records = self._client.read_test(id, fields)

        elif (self._app == 'bugzilla'):
            res, records = self._client.read(id, fields, query, limit, offset)
        elif (self._app == 'mantis'):
            res, records = self._client.read(id, fields, page, per_page)
        elif (self._app == 'trac'):
            res, records = self._client.read(id, fields, query)
        elif (self._app == 'jira'):
            res, records = self._client.read(id, fields, query, limit, offset)

        if (res):
            output = CommandlineTool.get_input_option('tr-output')
            if (not output):
                print(records)
            else:
                with (open(output, 'w')) as f:
                    f.write(str(records))
        else:
            print('Read error')

    def create(self):
        """Method create action

        Args:
           none

        Returns:
           void

        """

        if (self._app in ['qc', 'testlink']):
            params = {}
            if (self._entity in self._client._default_values):
                params = self._client._default_values[self._entity]
        else:
            params = self._client.default_values

        params_in = CommandlineTool.get_input_option('tr-params')
        if (params_in != False):
            params_in = params_in.split(',')
            for param in params_in:
                rec = param.split(':')
                key = rec[0]
                params[key] = rec[1]

        input = CommandlineTool.get_input_option('tr-input')
        if (input != False and path.exists(input)):
            with open(input, 'r') as f:
                params['description'] = f.read()

        if ('required_fields' in self._cfg and self._cfg['required_fields'] != None):
            if (self._app == 'qc'):
                if (self._entity in self._cfg['required_fields'] and
                        self._cfg['required_fields'][self._entity] != None):
                    fields = self._cfg['required_fields'][
                        self._entity].split(',')
                else:
                    fields = []
            else:
                fields = self._cfg['required_fields'].split(',')

            lov = {}
            if ('lov' in self._cfg and self._cfg['lov'] != None):
                if (self._app == 'qc'):
                    if (self._entity in self._cfg['lov'] and
                            self._cfg['lov'][self._entity] != None):
                        lov = {}
                        expr = self._cfg['lov'][self._entity]
                        if (expr != None):
                            lov_params = expr.split('#')
                            for param in lov_params:
                                conf = param.split('%')
                                lov[conf[0]] = conf[1]
                else:
                    lov = self._cfg['lov']

                if (len(lov) > 0):
                    for key, value in lov.items():
                        lov[key] = value.split(',')

            for field in fields:

                if (field not in params):
                    print('Enter required field: {0}'.format(field))

                    if (field not in lov):
                        value = raw_input(':')
                        params[field] = value

                    else:
                        prompt = ''
                        values = lov[field]
                        cnt = len(values)
                        for i in range(0, cnt):
                            prompt += '{0} - {1}\n'.format(i + 1, values[i])

                        valid = False
                        while (not valid):
                            try:
                                idx = int(raw_input('{0}'.format(prompt))) - 1
                            except ValueError:
                                idx = -1

                            if (idx >= 0 and idx < cnt):
                                valid = True
                            else:
                                print(
                                    'Wrong choice, enter required field: {0}'.format(field))

                        params[field] = values[idx]

        steps_in = CommandlineTool.get_input_option('tr-steps')
        steps = []
        if (steps_in != False):
            steps_in = steps_in.split('|')
            for step_in in steps_in:
                step = {}
                for param in step_in.split(','):
                    rec = param.split(':')
                    key = rec[0]
                    step[key] = rec[1]
                steps.append(step)

        if (self._app == 'qc'):
            if (self._entity in ('defect', 'test-instance')):
                id = self._client.create(self._entity, params)
            else:
                path = CommandlineTool.get_input_option('tr-path')
                if (not path):
                    print('Enter path')
                    path = raw_input(':')
                if (self._entity == 'test'):
                    id = self._client.create_test(path, params)
                elif (self._entity == 'test-set'):
                    id = self._client.create_test_set(path, params)
                else:
                    folders = path.split('/')
                    id = self._client.create_test_folder(
                        '/'.join(folders[:-1]), folders[-1], self._entity)

        elif (self._app == 'testlink'):
            if (self._entity == 'test-suite'):
                path = CommandlineTool.get_input_option('tr-path')
                if (not path):
                    print('Enter path')
                    path = raw_input(':')
                details = params['details'] if ('details' in params) else None

                folders = path.split('/')
                id = self._client.create_test_suite(
                    '/'.join(folders[:-1]), folders[-1], details)

            elif (self._entity == 'test-plan'):
                name = params['name'] if ('name' in params) else None
                if ('name' not in params):
                    print('Enter name')
                    name = raw_input(':')
                notes = params['notes'] if ('notes' in params) else None

                id = self._client.create_test_plan(name, notes)

            elif (self._entity == 'build'):
                plan = params['plan'] if ('plan' in params) else None
                if ('plan' not in params):
                    print('Enter plan id')
                    plan = raw_input(':')
                name = params['name'] if ('name' in params) else None
                if ('name' not in params):
                    print('Enter name')
                    name = raw_input(':')
                notes = params['notes'] if ('notes' in params) else None

                id = self._client.create_build(plan, name, notes)

            elif (self._entity == 'test'):
                path = CommandlineTool.get_input_option('tr-path')
                if (not path):
                    print('Enter path')
                    path = raw_input(':')

                id = self._client.create_test(path, params, steps)

        else:
            id = self._client.create(params)

        if (id != None):
            print('Record {0} created'.format(id))
        else:
            print('Create error')

    def update(self):
        """Method handles update action

        Args:
           none

        Returns:
           void

        """

        id = CommandlineTool.get_input_option('tr-id')
        if (not id):
            print('Enter id')
            id = raw_input(':')

        params_in = CommandlineTool.get_input_option('tr-params')
        if (not params_in):
            params = {}
        else:
            params_in = params_in.split(',')
            params = {}
            for param in params_in:
                rec = param.split(':')
                params[rec[0]] = rec[1]

        input = CommandlineTool.get_input_option('tr-input')
        if (input != False and path.exists(input)):
            with open(input, 'r') as f:
                params['description'] = f.read()

        if (self._app == 'qc'):
            res = self._client.update(id, self._entity, params)
        elif (self._app == 'testlink'):
            if (self._entity == 'test-plan'):
                test_id = params['test'] if ('test' in params) else None
                if ('test' not in params):
                    print('Enter test id')
                    test_id = raw_input(':')

                res = self._client.add_test_to_plan(test=test_id, plan_id=id)

            elif (self._entity == 'test'):
                plan_id = params['plan'] if ('plan' in params) else None
                if ('plan' not in params):
                    print('Enter plan id')
                    plan_id = raw_input(':')
                status = params['status'] if ('status' in params) else 'p'
                notes = params['notes'] if ('notes' in params) else None
                build_id = params['build'] if ('build' in params) else None

                res = self._client.update_test_execution(
                    id, status, notes, plan_id=plan_id, build_id=build_id)

        else:
            res = self._client.update(id, params)

        if (res):
            print('Record {0} updated'.format(id))
        else:
            print('Update error')

    def delete(self):
        """Method handles delete action

        Args:
           none

        Returns:
           void

        """

        id = CommandlineTool.get_input_option('tr-id')
        if (not id):
            print('Enter id')
            id = raw_input(':')

        if (self._app == 'qc'):
            res = self._client.delete(id, self._entity)
        else:
            res = self._client.delete(id)
        if (res):
            print('Record {0} deleted'.format(id))
        else:
            print('Delete error')



# -*- coding: utf-8 -*-
"""Client for TestLink

.. module:: trackapps.testlink
   :platform: Unix
   :synopsis: Client for TestLink
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
track_before_connect
track_after_connect
track_before_read
track_after_read
track_before_create
track_after_create
track_before_update
track_after_update
track_before_read_suite
track_after_rest_suite
track_before_read_plan
track_after_read_plan

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from hydratk.lib.network.rpc.client import RPCClient


class Client(object):
    """Class Client
    """

    _mh = None
    _client = None
    _url = None
    _dev_key = None
    _project = None
    _project_id = None
    _return_fields = None
    _default_values = {}
    _is_connected = None

    def __init__(self):
        """Class constructor

        Called when the object is initialized 

        Args:  
           none

        """

        self._mh = MasterHead.get_head()
        self._client = RPCClient('XMLRPC')

        cfg = self._mh.cfg['Extensions']['TrackApps']['testlink']
        if ('return_fields' in cfg and cfg['return_fields'] != None):
            self._return_fields = cfg['return_fields'].split(',')
        if ('default_values' in cfg and cfg['default_values'] != None):
            self._default_values = cfg['default_values']
        if ('url' in cfg and cfg['url'] != None):
            self._url = cfg['url']
        if ('dev_key' in cfg and cfg['dev_key'] != None):
            self._dev_key = cfg['dev_key']
        if ('project' in cfg and cfg['project'] != None):
            self._project = cfg['project']

    @property
    def client(self):
        """ client property getter """

        return self._client

    @property
    def url(self):
        """ url property getter """

        return self._url

    @property
    def dev_key(self):
        """ dev key property getter """

        return self._dev_key

    @property
    def project(self):
        """ project property getter """

        return self._project

    @property
    def project_id(self):
        """ project id property getter """

        return self._project_id

    @property
    def return_fields(self):
        """ return_fields property getter """

        return self._return_fields

    @property
    def default_values(self):
        """ default_values property getter """

        return self._default_values

    @property
    def is_connected(self):
        """ is_connected property getter """

        return self._is_connected

    def connect(self, url=None, dev_key=None, project=None):
        """Method connects to TestLink

        Args:    
           url (str): URL        
           dev_key (str): development key to access API
           project (str): project

        Returns:
           bool: result

        Raises:
           event: track_before_connect
           event: track_after_connect

        """

        message = 'url:{0}, dev_key:{1}, project:{2}'.format(
            url, dev_key, project)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_connecting', message), self._mh.fromhere())

        if (url == None):
            url = self._url
        if (dev_key == None):
            dev_key = self._dev_key
        if (project == None):
            project = self._project

        ev = event.Event('track_before_connect', url, dev_key, project)
        if (self._mh.fire_event(ev) > 0):
            url = ev.argv(0)
            dev_key = ev.argv(1)
            project = ev.argv(2)

        if (ev.will_run_default()):
            self._url = url
            self._dev_key = dev_key
            self._project = project

            url = self._url + '/testlink/lib/api/xmlrpc/v1/xmlrpc.php'
            params = {
                'testprojectname': self._project, 'devKey': self._dev_key}
            res = self._client.init_proxy(url)

        result = False
        if (res != None):
            body = self._client.call_method('tl.getTestProjectByName', params)

            if (body != None):
                if (body.__class__.__name__ == 'list'):
                    body = body[0]

                self._project_id = int(body['id']) if (
                    'id' in body.keys()) else None
                if (self._project_id != None):
                    self._is_connected = True
                    self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                        'track_connected'), self._mh.fromhere())
                    ev = event.Event('track_after_connect')
                    self._mh.fire_event(ev)
                    result = True
                else:
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', body['code'], body['message']), self._mh.fromhere())

        return result

    def read_test_suite(self, path, steps=True, fields=None):
        """Method reads tests under test suite

        Args: 
           path (str): suite path
           steps (bool): get steps
           fields (list): test fields to be returned, default all

        Returns:
           tuple: result (bool), records (dict), key - test suite, value - list of tests

        Raises:
           event: track_before_read_suite
           event: track_after_read_suite                           

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_reading_folder', path, 'test_suite'),
                      self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and self._return_fields != None):
            fields = self._return_fields

        ev = event.Event('track_before_read_suite', path, steps, fields)
        if (self._mh.fire_event(ev) > 0):
            path = ev.argv(0)
            steps = ev.argv(1)
            fields = ev.argv(2)

        if (ev.will_run_default()):

            id = self._get_suite(path)
            if (id == None):
                return False, None

            details = 'full' if (steps) else 'simple'
            params = {
                'devKey': self._dev_key, 'testsuiteid': id, 'details': details}
            recs = self._client.call_method(
                'tl.getTestCasesForTestSuite', params)
            if (recs == None):
                return False, None

            tests = {}
            suites = {str(id): {'name': path, 'parent': None, 'path': True}}
            for record in recs:
                parent = str(record['parent_id'])
                parent_orig = parent

                while (parent != None and parent not in suites):
                    params = {'devKey': self._dev_key, 'testsuiteid': parent}
                    rec = self._client.call_method(
                        'tl.getTestSuiteByID', params)
                    suites[parent] = {
                        'name': rec['name'], 'parent': str(rec['parent_id']), 'path': False}
                    parent = str(rec['parent_id'])

                for key, value in suites.items():
                    parent = None
                    while (not suites[key]['path']):
                        parent = suites[value['parent']] if (
                            parent == None) else parent['parent']
                        suites[key][
                            'name'] = '{0}/{1}'.format(parent['name'], suites[key]['name'])
                        if (parent['parent'] == None):
                            suites[key]['path'] = True

                suite_name = suites[parent_orig]['name']
                if (suite_name not in tests):
                    tests[suite_name] = []

                record_new = {}
                for key, value in record.items():
                    if (fields == None or key in fields):
                        record_new[key] = value

                if (record_new != {}):
                    tests[suite_name].append(record_new)

        if (recs != None):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_folder_read', len(recs)),
                          self._mh.fromhere())
            ev = event.Event('track_after_read_suite')
            self._mh.fire_event(ev)
            return True, tests

    def create_test_suite(self, path, name, details=None):
        """Method creates test folder on path

        Args: 
           path (str): suite path
           name (str): suite name
           details (str): suite details

        Returns:
           int: suite id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        params = 'path:{0}, name:{1}, details:{2}'.format(path, name, details)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'test_suite', params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        ev = event.Event('track_before_create', path, name, details)
        if (self._mh.fire_event(ev) > 0):
            path = ev.argv(0)
            name = ev.argv(1)
            details = ev.argv(2)

        if (ev.will_run_default()):
            suite = self._get_suite(path)
            if (suite == None):
                return None

            params = {'devKey': self._dev_key, 'testprojectid': self._project_id,
                      'testsuitename': name, 'parentid': suite, 'details': details}
            body = self._client.call_method('tl.createTestSuite', params)

        id = None
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            id = int(body['id']) if (
                'id' in body.keys() and int(body['id']) > 0) else None
            if (id != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_created', id), self._mh.fromhere())
                ev = event.Event('track_after_create')
                self._mh.fire_event(ev)
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', -1, body['msg']), self._mh.fromhere())

        return id

    def read_test_plan(self, plan=None, plan_id=None, build_id=None, fields=None):
        """Method reads tests under plan

        Args:
           plan (str): plan name, will be translated to plan_id
           plan_id (int): plan_id
           build_id (int): build id, latest build will be used if not specified
           fields (list): test fields to be returned, default all   

        Returns:
           tuple: result (bool), records (list of dict)

        Raises:
           event: track_before_read_plan
           event: track_after_read_plan                           

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_reading_folder', plan, 'test_plan'),
                      self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and self._return_fields != None):
            fields = self._return_fields

        ev = event.Event(
            'track_before_read_plan', plan, plan_id, build_id, fields)
        if (self._mh.fire_event(ev) > 0):
            plan = ev.argv(0)
            plan_id = ev.argv(1)
            build_id = ev.argv(2)
            fields = ev.argv(3)

        if (ev.will_run_default()):

            if (plan != None):
                params = {
                    'devKey': self._dev_key, 'testprojectname': self._project, 'testplanname': plan}
                body = self._client.call_method('tl.getTestPlanByName', params)
                if (body.__class__.__name__ == 'list'):
                    body = body[0]

                plan_id = body['id'] if ('id' in body.keys()) else None
                if (plan_id == None):
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', body['code'], body['message']), self._mh.fromhere())
                    return False, None

            if (build_id == None):
                params = {'devKey': self._dev_key, 'testplanid': plan_id}
                body = self._client.call_method(
                    'tl.getLatestBuildForTestPlan', params)
                if (body.__class__.__name__ == 'list'):
                    body = body[0]

                build_id = body['id'] if ('id' in body.keys()) else None
                if (build_id == None):
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', body['code'], body['message']), self._mh.fromhere())
                    return False, None

            params = {'devKey': self._dev_key,
                      'testplanid': plan_id, 'buildid': build_id}
            recs = self._client.call_method(
                'tl.getTestCasesForTestPlan', params)

            records_new = []
            if (len(recs) > 0):

                if (recs.__class__.__name__ == 'list'):
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', recs[0]['code'], recs[0]['message']), self._mh.fromhere())
                    return False, None

                records = []
                for key in sorted(recs.keys()):
                    records.append(recs[key][0])
                for record in records:
                    record_new = {}
                    for key, value in record.items():
                        if (fields == None or key in fields):
                            record_new[key] = value

                    if (record_new != {}):
                        records_new.append(record_new)

        if (recs != None):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_folder_read', len(records_new)),
                          self._mh.fromhere())
            ev = event.Event('track_after_read_plan')
            self._mh.fire_event(ev)
            return True, records_new
        else:
            return False, None

    def create_test_plan(self, name, notes=None):
        """Method creates test plan

        Args: 
           name (str): plan name
           notes (str): plan notes

        Returns:
           int: plan id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        message = 'name:{0}, notes:{1}'.format(name, notes)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'test_plan', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        ev = event.Event('track_before_create', name, notes)
        if (self._mh.fire_event(ev) > 0):
            name = ev.argv(0)
            notes = ev.argv(1)

        if (ev.will_run_default()):
            params = {'devKey': self._dev_key, 'testprojectname':
                      self._project, 'testplanname': name, 'notes': notes}
            body = self._client.call_method('tl.createTestPlan', params)

        id = None
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            id = int(body['id']) if (
                'id' in body.keys() and int(body['id']) > 0) else None
            if (id != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_created', id), self._mh.fromhere())
                ev = event.Event('track_after_create')
                self._mh.fire_event(ev)
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', body['code'], body['message']), self._mh.fromhere())

        return id

    def create_build(self, plan, name, notes=None):
        """Method creates build

        Args: 
           plan (int): plan id
           name (str): build name
           notes (str): build notes

        Returns:
           int: build id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        message = 'plan:{0}, name:{1}, notes:{1}'.format(plan, name, notes)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'build', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        ev = event.Event('track_before_create', plan, name, notes)
        if (self._mh.fire_event(ev) > 0):
            plan = ev.argv(0)
            name = ev.argv(1)
            notes = ev.argv(2)

        if (ev.will_run_default()):
            params = {'devKey': self._dev_key, 'testplanid': plan,
                      'buildname': name, 'buildnotes': notes}
            body = self._client.call_method('tl.createBuild', params)

        id = None
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            id = int(body['id']) if (
                'id' in body.keys() and int(body['id']) > 0) else None
            if (id != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_created', id), self._mh.fromhere())
                ev = event.Event('track_after_create')
                self._mh.fire_event(ev)
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', body['code'], body['message']), self._mh.fromhere())

        return id

    def read_test(self, id, fields=None):
        """Method reads test

        Args: 
           id (int): test id
           fields (list): fields to be returned default all

        Returns:
           tuple: result (bool), test (dict)

        Raises:
           event: track_before_read
           event: track_after_read

        """

        message = 'id:{0}, fields:{1}'.format(id, fields)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_reading', 'test', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        ev = event.Event('track_before_read', id, fields)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            fields = ev.argv(1)

        if (ev.will_run_default()):
            params = {'devKey': self._dev_key, 'testcaseid': id}
            body = self._client.call_method('tl.getTestCase', params)

        result, test = False, None
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            id = body['id'] if (
                'id' in body.keys() and int(body['id']) > 0) else None
            if (id != None):
                test = {}
                for key, value in body.items():
                    if (fields == None or key in fields):
                        test[key] = value

                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_read', 1), self._mh.fromhere())
                ev = event.Event('track_after_read')
                self._mh.fire_event(ev)
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', body['code'], body['message']), self._mh.fromhere())

        return result, test

    def create_test(self, path, params={}, steps=[]):
        """Method creates test

        Args: 
           path (str): suite path
           params (dict): params, key - field name, value - field value
           steps (list): list of dict, key - field name, value - field value

        Returns:
           int: test id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        message = 'path:{0}, params:{1}, steps:{2}'.format(path, params, steps)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'test', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        if (self._default_values != None):
            for key, value in self._default_values.items():
                if (key not in params):
                    params[key] = value

        ev = event.Event('track_before_create', path, params, steps)
        if (self._mh.fire_event(ev) > 0):
            path = ev.argv(0)
            params = ev.argv(1)
            steps = ev.argv(2)

        if (ev.will_run_default()):
            suite = self._get_suite(path)
            if (suite == None):
                return None

            for i in range(0, len(steps)):
                if ('step_number' not in steps[i].keys()):
                    steps[i]['step_number'] = i + 1

            params['devKey'] = self._dev_key
            params['testprojectid'] = self._project_id
            params['testsuiteid'] = suite
            params['steps'] = steps
            body = self._client.call_method('tl.createTestCase', params)

        id = None
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            id = int(body['id']) if (
                'id' in body.keys() and int(body['id']) > 0) else None
            if (id != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_created', id), self._mh.fromhere())
                ev = event.Event('track_after_create')
                self._mh.fire_event(ev)
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', body['code'], body['message']), self._mh.fromhere())

        return id

    def add_test_to_plan(self, test, plan=None, plan_id=None):
        """Method adds test to plan

        Args:            
           test (int): test id
           plan (str): plan name
           plan_id (int): plan_id

        Returns:
           bool: result

        Raises:
           event: track_before_update
           event: track_after_update

        """

        message = 'plan:{0}, plan_id:{1}'.format(plan, plan_id)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_updating', 'test_plan', test, message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_update', test, plan, plan_id)
        if (self._mh.fire_event(ev) > 0):
            test = ev.argv(0)
            plan = ev.argv(1)
            plan_id = ev.argv(2)

        if (ev.will_run_default()):

            if (plan != None):
                params = {
                    'devKey': self._dev_key, 'testprojectname': self._project, 'testplanname': plan}
                body = self._client.call_method('tl.getTestPlanByName', params)

                if (body.__class__.__name__ == 'list'):
                    body = body[0]

                plan_id = body['id'] if ('id' in body.keys()) else None
                if (plan_id == None):
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', body['code'], body['message']), self._mh.fromhere())
                    return False, None

            params = {'devKey': self._dev_key, 'testprojectid':
                      self._project_id, 'testplanid': plan_id, 'testcaseid': test}
            res, test_detail = self.read_test(test)
            if (res):
                params['version'] = int(test_detail['version'])
            body = self._client.call_method('tl.addTestCaseToTestPlan', params)

        result = False
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            status = body['status'] if ('status' in body.keys()) else None
            if (status != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_updated', test), self._mh.fromhere())
                ev = event.Event('track_after_update')
                self._mh.fire_event(ev)
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', body['code'], body['message']), self._mh.fromhere())

        return result

    def update_test_execution(self, test, status, notes=None, plan=None, plan_id=None, build_id=None):
        """Method updates test execution

        Args:
           test (int): test id
           status (str): status, p|f|b (passed|failed|blocked)
           notes (str): execution notes                
           plan (str): plan name, will be translated to id
           plan_id (int): plan id 
           build_id (int): build id, latest build will be used if not specified   

        Returns:
           bool: result   

        """

        message = 'status:{0}, notes:{1}, plan:{2}, plan_id:{3}, build_id:{4}'.format(
            status, notes, plan, plan_id, build_id)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_updating', 'test', test, message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event(
            'track_before_update', test, status, notes, plan, plan_id, build_id)
        if (self._mh.fire_event(ev) > 0):
            test = ev.argv(0)
            status = ev.argv(1)
            notes = ev.argv(2)
            plan = ev.argv(3)
            plan_id = ev.argv(4)
            build_id = ev.argv(5)

        if (ev.will_run_default()):

            if (plan != None):
                params = {
                    'devKey': self._dev_key, 'testprojectname': self._project, 'testplanname': plan}
                body = self._client.call_method('tl.getTestPlanByName', params)

                if (body.__class__.__name__ == 'list'):
                    body = body[0]

                plan_id = body['id'] if ('id' in body.keys()) else None
                if (plan_id == None):
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', body['code'], body['message']), self._mh.fromhere())
                    return False

            if (build_id == None):
                params = {'devKey': self._dev_key, 'testplanid': plan_id}
                body = self._client.call_method(
                    'tl.getLatestBuildForTestPlan', params)
                if (body.__class__.__name__ == 'list'):
                    body = body[0]

                build_id = body['id'] if ('id' in body.keys()) else None
                if (build_id == None):
                    self._mh.demsg('htk_on_error', self._mh._trn.msg(
                        'track_error', body['code'], body['message']), self._mh.fromhere())
                    return False

            params = {'devKey': self._dev_key, 'testplanid': plan_id, 'buildid':
                      build_id, 'testcaseid': test, 'status': status, 'notes': notes}
            body = self._client.call_method('tl.reportTCResult', params)

        result = False
        if (body != None):
            if (body.__class__.__name__ == 'list'):
                body = body[0]

            status = body['status'] if ('status' in body.keys()) else None
            if (status != None):
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_updated', test), self._mh.fromhere())
                ev = event.Event('track_after_update')
                self._mh.fire_event(ev)
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_error', body['code'], body['message']), self._mh.fromhere())

        return result

    def _get_suite(self, path):
        """Method gets suite from hierarchical path

        Args: 
           path (str): suite path, Suite1/Suite2/...    

        Returns:
           int: suite id

        """

        folders = path.split('/')
        cnt = len(folders)

        found = False
        id = None
        for i in range(0, cnt):
            if (i == 0):
                params = {
                    'devKey': self._dev_key, 'testprojectid': self._project_id}
                recs = self._client.call_method(
                    'tl.getFirstLevelTestSuitesForTestProject', params)
            else:
                params = {'devKey': self._dev_key, 'testsuiteid': id}
                recs = self._client.call_method(
                    'tl.getTestSuitesForTestSuite', params)

            found = False
            if (recs != None):
                if (recs.__class__.__name__ == 'dict'):
                    recs = [recs] if (
                        list(recs.values())[0].__class__.__name__ != 'dict') else recs.values()
                for rec in recs:
                    if (rec['name'] == folders[i]):
                        id = rec['id']
                        found = True
                        break

            if (not found):
                self._mh.demsg('htk_on_error', self._mh._trn.msg('track_unknown_folder', folders[i]),
                              self._mh.fromhere())
                return None

        return id



# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.testenv.translation.en
   :platform: Unix
   :synopsis: English language translation for TestEnv extension
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

language = {
    'name': 'English',
    'ISO-639-1': 'en'
}

msg = {
    'track_connecting': ["Connecting to server with params: {0}"],
    'track_connected': ["Connected successfully"],
    'track_missing_cookie': ["Cookie missing in response"],
    'track_missing_token': ["Token missing in response"],
    'track_missing_project': ["Project {0} not found"],
    'track_error': ["Error occured status:'{0}', error: {1}"],
    'track_disconnecting': ["Disconnecting from server"],
    'track_disconnected': ["Disconnected from server"],
    'track_not_connected': ["Not connected to server"],
    'track_reading': ["Reading records with params: {0}"],
    'track_read': ["Records read: '{0}'"],
    'track_creating': ["Creating record entity:'{0}', params: {1}"],
    'track_created': ["Record with id:'{0}' created"],
    'track_updating': ["Updating record entity:'{0}', id:'{1}', params: {2}"],
    'track_updated': ["Record with id:'{0}' updated"],
    'track_deleting': ["Deleting record entity:'{0}', id:'{1}'"],
    'track_deleted': ["Record with id:'{0}' deleted"],
    'track_unknown_record': ["Unknown record with id:'{0}'"],
    'track_unknown_entity': ["Unknown entity:'{0}'"],
    'track_unknown_folder': ["Unknown test folder:'{0}'"],
    'track_reading_folder': ["Reading test folder: '{0}', entity:'{1}'"],
    'track_folder_read': ["Tests in folder read: '{0}'"],
    'track_missing_hier': ["Missing hierarchical path in test folder:'{0}'"],
    'track_reading_set': ["Reading test set: '{0}'"],
    'track_set_read': ["Tests in set read: '{0}'"]
}



# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.trackapps.translation.cs
   :platform: Unix
   :synopsis: Czech language translation for TrackApps extension
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""
language = {
    'name': 'Čeština',
    'ISO-639-1': 'cs'
}


msg = {
    'track_connecting': ["Připojuji se ns server s parametry: {0}"],
    'track_connected': ["Spojení se serverem bylo úspěšné"],
    'track_missing_cookie': ["V odpovědi se nevrátilo cookie"],
    'track_missing_token': ["V odpovědi se nevrátil token"],
    'track_missing_project': ["Projekt {0} nenalezen"],
    'track_error': ["Nastala chyba status:'{0}', chyba: {1}"],
    'track_disconnecting': ["Ukončuji spojení se serverem"],
    'track_disconnected': ["Spojení se serverem bylo ukončeno"],
    'track_not_connected': ["Není navázáno spojení se serverem"],
    'track_reading': ["Čtu záznamy s parametry: {0}"],
    'track_read': ["Přečteno záznamů: '{0}'"],
    'track_creating': ["Vytvářím záznam entita:'{0}', parametry: {1}"],
    'track_created': ["Záznam s id:'{0}' vytvořen"],
    'track_updating': ["Měním záznam entita:'{0}', id:'{1}', parametry: {2}"],
    'track_updated': ["Záznam s id:'{0}' změněn"],
    'track_deleting': ["Mažu záznam entita:'{0}', id:'{1}'"],
    'track_deleted': ["Záznam s id:'{0}' smazán"],
    'track_unknown_record': ["Neexistující záznam s id:'{0}'"],
    'track_unknown_entity': ["Neexistující entita:'{0}'"],
    'track_unknown_folder': ["Neexistující test adresář:'{0}'"],
    'track_reading_folder': ["Čtu test adresář: '{0}', entita:'{1}'"],
    'track_folder_read': ["Přečteno testů v adresáři: '{0}'"],
    'track_missing_hier': ["Chybí hierarchical path v test adresáři:'{0}'"],
    'track_reading_set': ["Čtu test set: '{0}'"],
    'track_set_read': ["Přečteno testů v test setu: '{0}'"]
}



# -*- coding: utf-8 -*-
"""Client for Mantis

.. module:: trackapps.mantis
   :platform: Unix
   :synopsis: Client for Mantis
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
track_before_connect
track_after_connect
track_before_read
track_after_read
track_before_create
track_after_create
track_before_update
track_after_update
track_before_delete
track_after_delete

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from hydratk.lib.network.soap.client import SOAPClient
from lxml.etree import Element, SubElement, tostring

config = {
    'wsdl': '/api/soap/mantisconnect.php?wsdl',
    'ns': '{http://futureware.biz/mantisconnect}'
}

rec_fields = {
    'id': 'standard',
    'view_state': 'object_ref',
    'last_updated': 'date',
    'project': 'object_ref',
    'category': 'standard',
    'priority': 'object_ref',
    'severity': 'object_ref',
    'status': 'object_ref',
    'reporter': 'account_data',
    'summary': 'standard',
    'version': 'standard',
    'build': 'standard',
    'platform': 'standard',
    'os': 'standard',
    'os_build': 'standard',
    'reproducibility': 'object_ref',
    'date_submitted': 'date',
    'sponsorship_total': 'standard',
    'handler': 'account_data',
    'projection': 'object_ref',
    'eta': 'object_ref',
    'resolution': 'object_ref',
    'fixed_in_version': 'standard',
    'target_version': 'standard',
    'description': 'standard',
    'steps_to_reproduce': 'standard',
    'additional_information': 'standard',
    'attachments': 'attachment_data_array',
    'relationships': 'relationship_data_array',
    'due_date': 'date',
    'monitors': 'account_data_array',
    'sticky': 'standard',
    'tags': 'object_ref_array'
}


class Client(object):
    """Class Client
    """

    _mh = None
    _client = None
    _url = None
    _user = None
    _passw = None
    _project = None
    _project_id = None
    _return_fields = None
    _default_values = {}
    _is_connected = None

    def __init__(self):
        """Class constructor

        Called when the object is initialized 

        Args: 
           none

        """

        self._mh = MasterHead.get_head()
        self._client = SOAPClient()

        cfg = self._mh.cfg['Extensions']['TrackApps']['mantis']
        if ('return_fields' in cfg and cfg['return_fields'] != None):
            self._return_fields = cfg['return_fields'].split(',')
        if ('default_values' in cfg and cfg['default_values'] != None):
            self._default_values = cfg['default_values']
        if ('url' in cfg and cfg['url'] != None):
            self._url = cfg['url']
        if ('user' in cfg and cfg['user'] != None):
            self._user = cfg['user']
        if ('passw' in cfg and cfg['passw'] != None):
            self._passw = cfg['passw']
        if ('project' in cfg and cfg['project'] != None):
            self._project = cfg['project']

    @property
    def client(self):
        """ client property getter """

        return self._client

    @property
    def url(self):
        """ url property getter """

        return self._url

    @property
    def user(self):
        """ user property getter """

        return self._user

    @property
    def passw(self):
        """ passw property getter """

        return self._passw

    @property
    def project(self):
        """ project property getter """

        return self._project

    @property
    def project_id(self):
        """ project_id property getter """

        return self._project_id

    @property
    def return_fields(self):
        """ return_fields property getter """

        return self._return_fields

    @property
    def default_values(self):
        """ default_values property getter """

        return self._default_values

    @property
    def is_connected(self):
        """ is_connected property getter """

        return self._is_connected

    def connect(self, url=None, user=None, passw=None, project=None):
        """Method connects to Mantis

        Args:    
           url (str): URL        
           user (str): username
           passw (str): password
           project (str): QC project

        Returns:
           bool: result

        Raises:
           event: track_before_connect
           event: track_after_connect

        """

        message = 'url:{0}, user:{1}, passw:{2}, project:{3}'.format(
            url, user, passw, project)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_connecting', message), self._mh.fromhere())

        if (url == None):
            url = self._url
        if (user == None):
            user = self._user
        if (passw == None):
            passw = self._passw
        if (project == None):
            project = self._project

        ev = event.Event('track_before_connect', url, user, passw, project)
        if (self._mh.fire_event(ev) > 0):
            url = ev.argv(0)
            user = ev.argv(1)
            passw = ev.argv(2)
            project = ev.argv(3)

        if (ev.will_run_default()):
            self._url = url
            self._user = user
            self._passw = passw
            self._project = project

            url = self._url + config['wsdl']
            res = self._client.load_wsdl(url)

        if (res):

            root = Element(config['ns'] + 'mc_project_get_id_from_name')
            SubElement(root, 'username').text = self._user
            SubElement(root, 'password').text = self._passw
            SubElement(root, 'project_name').text = self._project

            res = self._client.send_request(
                'mc_project_get_id_from_name', body=tostring(root))
            if (res != None and res != 0):
                self._project_id = res
                self._is_connected = True
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_connected'), self._mh.fromhere())
                ev = event.Event('track_after_connect')
                self._mh.fire_event(ev)
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_missing_project', project), self._mh.fromhere())
                res = False

        return res

    def read(self, id=None, fields=None, page=-1, per_page=-1):
        """Method reads records

        Args: 
           id (int): record id         
           fields (list): fields to be returned, default all
           page (int): page number
           per_page (int): records per page

        Returns:
           tuple: result (bool), records (list of dictionary)

        Raises:
           event: track_before_read
           event: track_after_read

        """

        message = 'id:{0}, fields:{1}, page:{2}, per_page:{3}'.format(
            id, fields, page, per_page)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_reading', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and self._return_fields != None):
            fields = self._return_fields
        elif (fields == 'all'):
            fields = None

        ev = event.Event('track_before_read', id, fields, page, per_page)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            fields = ev.argv(1)
            page = ev.argv(2)
            per_page = ev.argv(3)

        if (ev.will_run_default()):

            if (id != None):
                root = Element(config['ns'] + 'mc_issue_get')
                SubElement(root, 'username').text = self._user
                SubElement(root, 'password').text = self._passw
                SubElement(root, 'issue_id').text = str(id)
                res = self._client.send_request(
                    'mc_issue_get', body=tostring(root))
            else:
                root = Element(config['ns'] + 'mc_project_get_issues')
                SubElement(root, 'username').text = self._user
                SubElement(root, 'password').text = self._passw
                SubElement(root, 'project_id').text = str(self._project_id)
                SubElement(root, 'page_number').text = str(page)
                SubElement(root, 'per_page').text = str(per_page)
                res = self._client.send_request(
                    'mc_project_get_issues', body=tostring(root))

            result = False
            records = None
            if (res != None):
                if (id != None):
                    parsed = self._parse_record(res, fields)
                    if (parsed != {}):
                        records = parsed
                    cnt = 1
                else:
                    records = []
                    for item in res:
                        parsed = self._parse_record(item, fields)
                        if (parsed != {}):
                            records.append(parsed)
                    cnt = len(records)
                result = True

                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_read', cnt), self._mh.fromhere())
                ev = event.Event('track_after_read')
                self._mh.fire_event(ev)

            return (result, records)

    def create(self, params):
        """Method creates record

        Args: 
           params (dict): record content, key - field name, value - field value

        Returns:
           int: record id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'issue', params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        if (self._default_values != {}):
            for key, value in self._default_values.items():
                if (key not in params):
                    params[key] = value

        ev = event.Event('track_before_create', params)
        if (self._mh.fire_event(ev) > 0):
            params = ev.argv(0)

        if (ev.will_run_default()):

            root = Element(config['ns'] + 'mc_issue_add')
            SubElement(root, 'username').text = self._user
            SubElement(root, 'password').text = self._passw
            root.append(self._toxml(params))

            res = self._client.send_request(
                'mc_issue_add', body=tostring(root))

        id = None
        if (res != None):
            id = int(res)
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_created', id), self._mh.fromhere())
            ev = event.Event('track_after_create')
            self._mh.fire_event(ev)

        return id

    def update(self, id, params={}):
        """Method updates record

        Args: 
           id (int): record id       
           params (dict): record content, key - field name, value - field value 

        Returns:
           bool: result

        Raises:
           event: track_before_update
           event: track_after_update

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_updating', 'issue', id, params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_update', id, params)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            params = ev.argv(1)

        if (ev.will_run_default()):

            record = self.read(id, fields='all')[1]
            if (record == None or len(record) == 0):
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_unknown_record', id), self._mh.fromhere())
                return None
            params_old = record
            for key, value in params.items():
                params_old[key] = value

            root = Element(config['ns'] + 'mc_issue_update')
            SubElement(root, 'username').text = self._user
            SubElement(root, 'password').text = self._passw
            SubElement(root, 'issueId').text = str(id)
            root.append(self._toxml(params_old))

            res = self._client.send_request(
                'mc_issue_update', body=tostring(root))

        result = False
        if (res != None):
            result = True
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_updated', id), self._mh.fromhere())
            ev = event.Event('track_after_update')
            self._mh.fire_event(ev)

        return result

    def delete(self, id):
        """Method deletes record

        Args: 
           id (int): record id       

        Returns:
           bool: result

        Raises:
           event: track_before_delete
           event: track_after_delete

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_deleting', 'issue', id), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_delete', id)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)

        if (ev.will_run_default()):

            root = Element(config['ns'] + 'mc_issue_delete')
            SubElement(root, 'username').text = self._user
            SubElement(root, 'password').text = self._passw
            SubElement(root, 'issue_id').text = str(id)

            res = self._client.send_request(
                'mc_issue_delete', body=tostring(root))

        result = False
        if (res == True):
            result = True
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_deleted', id), self._mh.fromhere())
            ev = event.Event('track_after_delete')
            self._mh.fire_event(ev)

        return result

    def _parse_record(self, rec, fields=None):
        """Method parses record

        Args: 
           rec (xml): record in xml form
           fields (list): fields to be returned, default all 

        Returns:
           dict: parsed record

        """

        record = {}

        for key, value in rec_fields.items():
            if ((fields == None or key in fields) and hasattr(rec, key)):
                if (value == 'standard'):
                    record[key] = getattr(rec, key)
                elif (value == 'date'):
                    record[key] = str(getattr(rec, key))
                elif (value == 'object_ref'):
                    attr = getattr(rec, key)
                    id = attr.id if (hasattr(attr, 'id')) else None
                    name = attr.name if (hasattr(attr, 'name')) else None
                    record[key] = {'id': id, 'name': name}
                elif (value == 'object_ref_array'):
                    array = []
                    for item in getattr(rec, key):
                        id = item.id if (hasattr(item, 'id')) else None
                        name = item.name if (hasattr(item, 'name')) else None
                        array.append({'id': id, 'name': name})
                    record[key] = array
                elif (value == 'account_data'):
                    attr = getattr(rec, key)
                    id = attr.id if (hasattr(attr, 'id')) else None
                    name = attr.name if (hasattr(attr, 'name')) else None
                    real_name = attr.real_name if (
                        hasattr(attr, 'real_name')) else None
                    email = attr.email if (hasattr(attr, 'email')) else None
                    record[key] = {
                        'id': id, 'name': name, 'real_name': real_name, 'email': email}
                elif (value == 'account_data_array'):
                    array = []
                    for item in getattr(rec, key):
                        id = item.id if (hasattr(item, 'id')) else None
                        name = item.name if (hasattr(item, 'name')) else None
                        real_name = item.real_name if (
                            hasattr(item, 'real_name')) else None
                        email = item.email if (
                            hasattr(item, 'email')) else None
                        array.append(
                            {'id': id, 'name': name, 'real_name': real_name, 'email': email})
                elif (value == 'attachment_data_array'):
                    array = []
                    for item in getattr(rec, key):
                        id = item.id if (hasattr(item, 'id')) else None
                        filename = item.filename if (
                            hasattr(item, 'filename')) else None
                        size = item.size if (hasattr(item, 'size')) else None
                        download_url = item.download_url if (
                            hasattr(item, 'download_url')) else None
                        user_id = item.user_id if (
                            hasattr(item, 'user_id')) else None
                        array.append({'id': id, 'filename': filename, 'size': size,
                                      'content_type': item.content_type, 'date_submitted': item.date_submitted,
                                      'download_url': download_url, 'user_id': user_id})
                    record[key] = array
                elif (value == 'relationship_data_array'):
                    array = []
                    for item in getattr(rec, key):
                        id = item.id if (hasattr(item, 'id')) else None
                        type_id = item.type.id if (
                            hasattr(item, 'type') and hasattr(item.type, 'id')) else None
                        type_name = item.type.name if (
                            hasattr(item, 'type') and hasattr(item.type, 'name')) else None
                        target_id = item.target_id if (
                            hasattr(item, 'target_id')) else None
                        array.append({'id': id, 'type': {'id': type_id, 'name': type_name},
                                      'target_id': target_id})
                    record[key] = array

        return record

    def _toxml(self, params):
        """Method creates record in xml

        Args: 
           params (dict): record content, key - field name, value - field value

        Returns:
           xml: record

        """

        root = Element('issue')

        if ('project' not in params):
            elem = SubElement(root, 'project')
            SubElement(elem, 'id').text = str(self._project_id)
            SubElement(elem, 'name').text = self._project

        for key, value in params.items():
            type = rec_fields[key] if (key in rec_fields) else 'standard'

            if (type in ('standard', 'date')):
                SubElement(root, key).text = str(value)
            elif (type == 'object_ref'):
                elem = SubElement(root, key)
                SubElement(elem, 'id').text = str(
                    value['id']) if ('id' in value) else None
                SubElement(elem, 'name').text = value['name'].decode('utf8') if (
                    'name' in value and value['name'] != None) else None
            elif (type == 'object_ref_array'):
                elem = SubElement(root, key)
                for item in value:
                    el_item = Element('item')
                    SubElement(elem, 'id').text = item['id'] if (
                        'id' in item) else None
                    SubElement(elem, 'name').text = item['name'].decode('utf8') if (
                        'name' in item and item['name'] != None) else None
                    elem.append(item)
            elif (type == 'account_data'):
                elem = SubElement(root, key)
                SubElement(elem, 'id').text = str(
                    value['id']) if ('id' in value) else None
                SubElement(elem, 'name').text = value['name'].decode('utf8') if (
                    'name' in value and value['name'] != None) else None
                SubElement(elem, 'real_name').text = value['real_name'].decode('utf8') if (
                    'real_name' in value and value['real_name'] != None) else None
                SubElement(elem, 'email').text = value['email'].decode('utf8') if (
                    'email' in value and value['email'] != None) else None
            elif (type == 'account_data_array'):
                elem = SubElement(root, key)
                for item in value:
                    el_item = Element('item')
                    SubElement(elem, 'id').text = str(
                        item['id']) if ('id' in item) else None
                    SubElement(elem, 'name').text = item['name'].decode('utf8') if (
                        'name' in item and item['name'] != None) else None
                    SubElement(elem, 'real_name').text = item['real_name'].decode('utf8') if (
                        'real_name' in item and item['real_name'] != None) else None
                    SubElement(elem, 'email').text = item['email'].decode('utf8') if (
                        'email' in item and item['email'] != None) else None
                    elem.append(item)
            elif (type == 'attachment_data_array'):
                elem = SubElement(root, key)
                for item in value:
                    el_item = Element('item')
                    SubElement(elem, 'id').text = str(
                        item['id']) if ('id' in item) else None
                    SubElement(elem, 'filename').text = item['filename'].decode('utf8') if (
                        'filename' in item and item['filename'] != None) else None
                    SubElement(elem, 'size').text = item['size'] if (
                        'size' in item) else None
                    SubElement(elem, 'download_url').text = item['download_url'] if (
                        'download_url' in item) else None
                    SubElement(elem, 'user_id').text = item['user_id'] if (
                        'user_id' in item) else None
                    elem.append(item)
            elif (type == 'relationship_data_array'):
                elem = SubElement(root, key)
                for item in value:
                    el_item = Element('item')
                    SubElement(elem, 'id').text = str(
                        item['id']) if ('id' in item) else None
                    if ('type' in item):
                        el_type = SubElement(el_item)
                        SubElement(el_type, 'id').text = str(
                            item['type']['id']) if ('id' in item['type']) else None
                        SubElement(el_type, 'name').text = item['type']['name'].decode('utf8') if (
                            'name' in item['type'] and item['type']['name'] != None) else None
                    SubElement(elem, 'target_id').text = item['target_id'] if (
                        'target_id' in item) else None
                    elem.append(item)

        return root



# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.trackapps.translation.cs.help
   :platform: Unix
   :synopsis: Czech language translation for TrackAPps extension help generator
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""
language = {
    'name': 'Čeština',
    'ISO-639-1': 'cs'
}

''' TrackApps Commands '''
help_cmd = {
    'track': 'spustit konzolové rozšíření trackapps',

    # standalone with option profile trackapps
    'run': 'spustit konzolové rozšíření trackapps'
}

''' TrackApps Options '''
help_opt = {
    'tr-app': {'{h}--tr-app qc|bugzilla|mantis|trac|testlink{e}': {'description': 'aplikace', 'commands': ('track')}},
    'tr-action': {'{h}--tr-action read|create|update|delete{e}': {'description': 'akce, delete podporováno pro aplikace: qc|mantis|trac', 'commands': ('track')}},
    'tr-type': {'[{h}--tr-type defect|test-folder|test|test-set-folder|test-set|test-instance|test-suite|test-plan|build{e}]': {'description': 'typ entity, default defect, podporováno pro akce: read|create|update|delete, aplikace: qc|testlink', 'commands': ('track')}},
    'tr-input': {'[{h}--tr-input <soubor>{e}]': {'description': 'jméno souboru, obsah se zapíše do description tiketu, podporováno pro akce: create|update', 'commands': ('track')}},
    'tr-output': {'[{h}--tr-output <soubor>{e}]': {'description': 'jméno souboru, zápis výstupu akce, podporováno pro akci: read', 'commands': ('track')}},
    'tr-url': {'[{h}--tr-url <url>{e}]': {'description': 'url, konfigurovatelné', 'commands': ('track')}},
    'tr-user': {'[{h}--tr-user <uživatel>{e}]': {'description': 'uživatelské jméno, konfigurovatelné', 'commands': ('track')}},
    'tr-passw': {'[{h}--tr-passw <heslo>{e}]': {'description': 'heslo, konfigurovatelné', 'commands': ('track')}},
    'tr-dev-key': {'[{h}--tr-dev-key <klíč>{e}]': {'description': 'klíč vývojáře, konfigurovatelné, podporováno pro aplikaci: testlink', 'commands': ('track')}},
    'tr-domain': {'[{h}--tr-domain <doména>{e}]': {'description': 'doména, konfigurovatelné, podporováno pro aplikaci: qc', 'commands': ('track')}},
    'tr-project': {'[{h}--tr-project <projekt>{e}]': {'description': 'projekt, konfigurovatelné, podporováno pro aplikace: qc|mantis|trac|jira|testlink', 'commands': ('track')}},
    'tr-id': {'[{h}--tr-id <číslo>{e}]': {'description': 'id záznamu, podporováno pro akce: read|update|delete', 'commands': ('track')}},
    'tr-fields': {'[{h}--tr-fields <seznam>{e}]': {'description': 'požadovaná pole, jméno1,jméno2,... , podporováno pro akci: read', 'commands': ('track')}},
    'tr-query': {'[{h}--tr-query <výraz>{e}]': {'description': 'dotaz, podporováno pro akci: read, aplikace: qc|bugzilla|trac|jira', 'commands': ('track')}},
    'tr-order-by': {'[{h}--tr-order-by <výraz>{e}]': {'description': 'řazení záznamů, jméno1:směr,jméno2:směr,... , směr asc|desc, podporováno pro akci: read, aplikaci: qc', 'commands': ('track')}},
    'tr-limit': {'[{h}--tr-limit <číslo>{e}]': {'description': 'limit, podporováno pro akci: read, aplikace: qz|bugzilla|jira', 'commands': ('track')}},
    'tr-offset': {'[{h}--tr-offset <číslo>{e}]': {'description': 'offset, podporováno pro akci: read, aplikace: qc|bugzilla|jira', 'commands': ('track')}},
    'tr-page': {'[{h}--tr-page <číslo>{e}]': {'description': 'stránka záznamů, podporováno pro akci: read, aplikaci: mantis', 'commands': ('track')}},
    'tr-per-page': {'[{h}--tr-per-page <číslo>{e}]': {'description': 'počet záznamů na stránku, podporováno pro akci read: aplikaci: mantis', 'commands': ('track')}},
    'tr-params': {'[{h}--tr-params <slovník>{e}]': {'description': 'parametry záznamu, jméno1:hodnota,jméno2:hodnota, podporováno pro akce: create|update', 'commands': ('track')}},
    'tr-path': {'[{h}--tr-path <cesta>{e}]': {'description': 'adresářová cesta, dir1/dir2/... , podporováno pro případy: read/create folder|read/create test set|create test|read/create suite, aplikace: qc|testlink', 'commands': ('track')}},
    'tr-steps': {'[{h}--tr-steps <seznam>{e}]': {'description': 'kroky testu oddělené |, parametry kroku používají tvar slovníku, jméno1:hodnota,jméno2:hodnota,...|jméno1:hodnota,jméno2:hodnota,... , podporováno akci: create, aplikaci: testlink', 'commands': ('track')}},

    # standalone with option profile trackapps
    'app': {'{h}--app qc|bugzilla|mantis|trac|testlink{e}': {'description': 'aplikace', 'commands': ('run')}},
    'action': {'{h}--action read|create|update|delete{e}': {'description': 'akce, delete podporováno pro aplikace: qc|mantis|trac', 'commands': ('run')}},
    'type': {'[{h}--type defect|test-folder|test|test-set-folder|test-set|test-instance|test-suite|test-plan|build{e}]': {'description': 'typ entity, default defect, podporováno pro akce: read|create|update|delete, aplikace: qc|testlink', 'commands': ('run')}},
    'input': {'[{h}--input <soubor>{e}]': {'description': 'jméno souboru, obsah se zapíše do description tiketu, podporováno pro akce: create|update', 'commands': ('run')}},
    'output': {'[{h}--output <soubor>{e}]': {'description': 'jméno souboru, zápis výstupu akce, podporováno pro akci: read', 'commands': ('run')}},
    'url': {'[{h}--url <url>{e}]': {'description': 'url, konfigurovatelné', 'commands': ('run')}},
    'user': {'[{h}--user <uživatel>{e}]': {'description': 'uživatelské jméno, konfigurovatelné', 'commands': ('run')}},
    'passw': {'[{h}--passw <heslo>{e}]': {'description': 'heslo, konfigurovatelné', 'commands': ('run')}},
    'dev-key': {'[{h}--dev-key <klíč>{e}]': {'description': 'klíč vývojáře, konfigurovatelné, podporováno pro aplikaci: testlink', 'commands': ('run')}},
    'domain': {'[{h}--domain <doména>{e}]': {'description': 'doména, konfigurovatelné, podporováno pro aplikaci: qc', 'commands': ('run')}},
    'project': {'[{h}--project <projekt>{e}]': {'description': 'projekt, konfigurovatelné, podporováno pro aplikace: qc|mantis|trac|jira|testlink', 'commands': ('run')}},
    'id': {'[{h}--id <číslo>{e}]': {'description': 'id záznamu, podporováno pro akce: read|update|delete', 'commands': ('run')}},
    'fields': {'[{h}--fields <seznam>{e}]': {'description': 'požadovaná pole, jméno1,jméno2,... , podporováno pro akci: read', 'commands': ('run')}},
    'query': {'[{h}--query <výraz>{e}]': {'description': 'dotaz, podporováno pro akci: read, aplikace: qc|bugzilla|trac|jira', 'commands': ('run')}},
    'order-by': {'[{h}--order-by <výraz>{e}]': {'description': 'řazení záznamů, jméno1:směr,jméno2:směr,... , směr asc|desc, podporováno pro akci: read, aplikaci: qc', 'commands': ('run')}},
    'limit': {'[{h}--limit <číslo>{e}]': {'description': 'limit, podporováno pro akci: read, aplikace: qz|bugzilla|jira', 'commands': ('run')}},
    'offset': {'[{h}--offset <číslo>{e}]': {'description': 'offset, podporováno pro akci: read, aplikace: qc|bugzilla|jira', 'commands': ('run')}},
    'page': {'[{h}--page <číslo>{e}]': {'description': 'stránka záznamů, podporováno pro akci: read, aplikaci: mantis', 'commands': ('run')}},
    'per-page': {'[{h}--per-page <číslo>{e}]': {'description': 'počet záznamů na stránku, podporováno pro akci read: aplikaci: mantis', 'commands': ('run')}},
    'params': {'[{h}--params <slovník>{e}]': {'description': 'parametry záznamu, jméno1:hodnota,jméno2:hodnota, podporováno pro akce: create|update', 'commands': ('run')}},
    'path': {'[{h}--path <cesta>{e}]': {'description': 'adresářová cesta, dir1/dir2/... , podporováno pro případy: read/create folder|read/create test set|create test|read/create suite, aplikace: qc|testlink', 'commands': ('run')}},
    'steps': {'[{h}--steps <seznam>{e}]': {'description': 'kroky testu oddělené |, parametry kroku používají tvar slovníku, jméno1:hodnota,jméno2:hodnota,...|jméno1:hodnota,jméno2:hodnota,... , podporováno akci: create, aplikaci: testlink', 'commands': ('run')}}
}



# -*- coding: utf-8 -*-
from setuptools import setup as st_setup
from setuptools import find_packages as st_find_packages
from sys import argv
import hydratk.lib.install.task as task
import hydratk.lib.system.config as syscfg

try:
    os_info = syscfg.get_supported_os()
except Exception as exc:
    print(str(exc))
    exit(1)

with open("README.rst", "r") as f:
    readme = f.read()

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Environment :: Other Environment",
    "Intended Audience :: Developers",
    "License :: Freely Distributable",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: Implementation",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Software Development :: Libraries :: Application Frameworks",
    "Topic :: Utilities"
]

config = {
    'pre_tasks': [
        task.install_modules
    ],

    'post_tasks': [
        task.set_config,
        task.set_manpage
    ],

    'modules': [
        {'module': 'hydratk', 'version': '>=0.5.0'},
        {'module': 'hydratk-lib-network', 'version': '>=0.2.0'}
    ],

    'files': {
        'config': {
            'etc/hydratk/conf.d/hydratk-ext-trackapps.conf': '{0}/hydratk/conf.d'.format(syscfg.HTK_ETC_DIR)
        },
        'manpage': 'doc/trackapps.1'
    }
}

task.run_pre_install(argv, config)

entry_points = {
    'console_scripts': [
        'trackapps = hydratk.extensions.trackapps.bootstrapper:run_app'
    ]
}

st_setup(
    name='hydratk-ext-trackapps',
    version='0.1.3',
    description='Interface to bugtracking and test management applications',
    long_description=readme,
    author='Petr Rašek, HydraTK team',
    author_email='bowman@hydratk.org, team@hydratk.org',
    url='http://extensions.hydratk.org/trackapps',
    license='BSD',
    packages=st_find_packages('src'),
    package_dir={'': 'src'},
    classifiers=classifiers,
    zip_safe=False,
    entry_points=entry_points,
    keywords='hydratk,Quality Center,Bugzilla,Mantis,Jira,Trac,TestLink',
    requires_python='>=2.6,!=3.0.*,!=3.1.*,!=3.2.*',
    platforms='Linux,FreeBSD'
)

task.run_post_install(argv, config)



# -*- coding: utf-8 -*-
"""Client for Trac

.. module:: trackapps.trac
   :platform: Unix
   :synopsis: Client for Trac
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
track_before_connect
track_after_connect
track_before_read
track_after_read
track_before_create
track_after_create
track_before_update
track_after_update
track_before_delete
track_after_delete

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from hydratk.lib.network.rpc.client import RPCClient
from sys import version_info


class Client(object):
    """Class Client
    """

    _mh = None
    _client = None
    _url = None
    _user = None
    _passw = None
    _project = None
    _return_fields = None
    _default_values = {}
    _is_connected = None

    def __init__(self):
        """Class constructor

        Called when the object is initialized   

        Args:
           none

        """

        self._mh = MasterHead.get_head()
        self._client = RPCClient('XMLRPC')

        cfg = self._mh.cfg['Extensions']['TrackApps']['trac']
        if ('return_fields' in cfg and cfg['return_fields'] != None):
            self._return_fields = cfg['return_fields'].split(',')
        if ('default_values' in cfg and cfg['default_values'] != None):
            self._default_values = cfg['default_values']
        if ('url' in cfg and cfg['url'] != None):
            self._url = cfg['url']
        if ('user' in cfg and cfg['user'] != None):
            self._user = cfg['user']
        if ('passw' in cfg and cfg['passw'] != None):
            self._passw = cfg['passw']
        if ('project' in cfg and cfg['project'] != None):
            self._project = cfg['project']

    @property
    def client(self):
        """ client property getter """

        return self._client

    @property
    def url(self):
        """ url property getter """

        return self._url

    @property
    def user(self):
        """ user property getter """

        return self._user

    @property
    def passw(self):
        """ passw property getter """

        return self._passw

    @property
    def project(self):
        """ project property getter """

        return self._project

    @property
    def return_fields(self):
        """ return_fields property getter """

        return self._return_fields

    @property
    def default_values(self):
        """ default_values property getter """

        return self._default_values

    @property
    def is_connected(self):
        """ is_connected property getter """

        return self._is_connected

    def connect(self, url=None, user=None, passw=None, project=None):
        """Method connects to Trac

        Args:    
           url (str): URL        
           user (str): username
           passw (str): password
           project (str): project

        Returns:
           bool: result

        Raises:
           event: track_before_connect
           event: track_after_connect

        """

        message = 'url:{0}, user:{1}, passw:{2}, project:{3}'.format(
            url, user, passw, project)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_connecting', message), self._mh.fromhere())

        if (url == None):
            url = self._url
        if (user == None):
            user = self._user
        if (passw == None):
            passw = self._passw
        if (project == None):
            project = self._project

        ev = event.Event('track_before_connect', url, user, passw, project)
        if (self._mh.fire_event(ev) > 0):
            url = ev.argv(0)
            user = ev.argv(1)
            passw = ev.argv(2)
            project = ev.argv(3)

        if (ev.will_run_default()):
            self._url = url
            self._user = user
            self._passw = passw
            self._project = project

            url = self._url
            if (self._user != None):
                if ('http://' in self._url):
                    url = 'http://{0}:{1}@{2}'.format(
                        self._user, self._passw, self._url.replace('http://', ''))
                elif ('https://' in self._url):
                    url = 'https://{0}:{1}@{2}'.format(
                        self._user, self._passw, self._url.replace('https://', ''))
                else:
                    url = '{0}:{1}@{2}'.format(
                        self._user, self._passw, self._url)

            url += '/{0}/login/xmlrpc'.format(self._project)
            res = self._client.init_proxy(url)

        result = False
        if (res and self._client.call_method('system.getAPIVersion')):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_connected'), self._mh.fromhere())
            ev = event.Event('track_after_connect')
            self._mh.fire_event(ev)
            self._is_connected = True
            result = True

        return result

    def read(self, id=None, fields=None, query=None):
        """Method reads records

        Args: 
           id (int): record id         
           fields (list): fields to be returned, default all
           query (str): query, see Trac doc

        Returns:
           tuple: result (bool), records (list of dictionary)

        Raises:
           event: track_before_read
           event: track_after_read

        """

        message = 'id:{0}, fields:{1}, query:{2}'.format(id, fields, query)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_reading', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and self._return_fields != None):
            fields = self._return_fields

        ev = event.Event('track_before_read', id, fields, query)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            fields = ev.argv(1)
            query = ev.argv(2)

        if (ev.will_run_default()):

            if (id != None):
                body = self._client.call_method('ticket.get', id)
                recs = [body] if (body != None) else []
            else:
                rec_id = self._client.call_method('ticket.query', query)
                recs = []
                for id in rec_id:
                    recs.append(self._client.call_method('ticket.get', id))

            result = False
            records = None

            if (len(recs) > 0):

                records = []
                for rec in recs:
                    record = {}

                    record['id'] = rec[0]
                    for key, value in rec[3].items():
                        if (fields == None or key in fields):
                            record[key] = str(value)

                    if (record != {}):
                        records.append(record)

                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_read', len(records)), self._mh.fromhere())
                ev = event.Event('track_after_read')
                self._mh.fire_event(ev)
                result = True

            return (result, records)

    def create(self, params={}):
        """Method creates record

        Args: 
           params (dict): record content, key - field name, value - field value

        Returns:
           int: record id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', 'issue', params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        if (self._default_values != {}):
            for key, value in self._default_values.items():
                if (key not in params):
                    params[key] = value

        ev = event.Event('track_before_create', params)
        if (self._mh.fire_event(ev) > 0):
            params = ev.argv(0)

        if (ev.will_run_default()):

            summary, description, attrs = None, None, {}
            for key, value in params.items():
                if (key == 'summary'):
                    summary = str(value)
                elif (key == 'description'):
                    description = str(value)
                else:
                    attrs[key] = str(value).decode('utf8') if (
                        version_info[0] == 2) else str(value)

        id = self._client.call_method(
            'ticket.create', summary, description, attrs)
        if (id != None):
            id = int(id)
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_created', id), self._mh.fromhere())
            ev = event.Event('track_after_create')
            self._mh.fire_event(ev)

        return id

    def update(self, id, params={}):
        """Method updates record

        Args: 
           id (int): record id 
           params (dict): record content, key - field name, value - field value

        Returns:
           bool: result

        Raises:
           event: track_before_update
           event: track_after_update

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_updating', 'issue', id, params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_update', id, params)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)
            params = ev.argv(1)

        if (ev.will_run_default()):

            attrs = {}
            for key, value in params.items():
                attrs[key] = str(value).decode('utf8') if (
                    version_info[0] == 2) else str(value)

            body = self._client.call_method(
                'ticket.update', id, 'comment', attrs)

        result = False
        if (body != None):
            result = True
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_updated', id), self._mh.fromhere())
            ev = event.Event('track_after_update')
            self._mh.fire_event(ev)

        return result

    def delete(self, id):
        """Method deletes record

        Args: 
           id (int): record id 

        Returns:
           bool: result

        Raises:
           event: track_before_delete
           event: track_after_delete

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_deleting', 'issue', id), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_delete', id)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)

        if (ev.will_run_default()):
            body = self._client.call_method('ticket.delete', id)

        result = False
        if (body != None):
            result = True
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_deleted', id), self._mh.fromhere())
            ev = event.Event('track_after_delete')
            self._mh.fire_event(ev)

        return result



# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.trackapps.translation.en.help
   :platform: Unix
   :synopsis: English language translation for TrackApps extension help generator
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""
language = {
    'name': 'English',
    'ISO-639-1': 'en'
}

''' TrackApps Commands '''
help_cmd = {
    'track': 'run trackapps command line extension',

    # standalone with option profile trackapps
    'run': 'run trackapps command line extension'
}

''' TrackApps Options '''
help_opt = {
    'tr-app': {'{h}--tr-app qc|bugzilla|mantis|trac|testlink{e}': {'description': 'application', 'commands': ('track')}},
    'tr-action': {'{h}--tr-action read|create|update|delete{e}': {'description': 'action, delete supported for apps: qc|mantis|trac', 'commands': ('track')}},
    'tr-type': {'[{h}--tr-type defect|test-folder|test|test-set-folder|test-set|test-instance|test-suite|test-plan|build{e}]': {'description': 'entity type, default defect, supported for actions: read|create|update|delete, apps: qc|testlink', 'commands': ('track')}},
    'tr-input': {'[{h}--tr-input <filename>{e}]': {'description': 'filename, content is written to ticket description, supported for actions: create|update', 'commands': ('track')}},
    'tr-output': {'[{h}--tr-output <filename>{e}]': {'description': 'filename, writes action output, supported for action: read', 'commands': ('track')}},
    'tr-url': {'[{h}--tr-url <url>{e}]': {'description': 'url, configurable', 'commands': ('track')}},
    'tr-user': {'[{h}--tr-user <username>{e}]': {'description': 'username, configurable', 'commands': ('track')}},
    'tr-passw': {'[{h}--tr-passw <password>{e}]': {'description': 'password, configurable', 'commands': ('track')}},
    'tr-dev-key': {'[{h}--tr-dev-key <key>{e}]': {'description': 'developer key, configurable, supported for app: testlink', 'commands': ('track')}},
    'tr-domain': {'[{h}--tr-domain <domain>{e}]': {'description': 'domain, configurable, supported for app: qc', 'commands': ('track')}},
    'tr-project': {'[{h}--tr-project <project>{e}]': {'description': 'project, configurable, supported for apps: qc|mantis|trac|jira|testlink', 'commands': ('track')}},
    'tr-id': {'[{h}--tr-id <num>{e}]': {'description': 'record id, supported for actions: read|update|delete', 'commands': ('track')}},
    'tr-fields': {'[{h}--tr-fields <list>{e}]': {'description': 'requested fields, name1,name2,... , supported for action: read', 'commands': ('track')}},
    'tr-query': {'[{h}--tr-query <expression>{e}]': {'description': 'query, supported for action: read, apps: qc|bugzilla|trac|jira', 'commands': ('track')}},
    'tr-order-by': {'[{h}--tr-order-by <expression>{e}]': {'description': 'record ordering, name1:direction,name2:direction,... , direction asc|desc, supported for action: read, app: qc', 'commands': ('track')}},
    'tr-limit': {'[{h}--tr-limit <num>{e}]': {'description': 'limit, supported for action: read, apps: qc|bugzilla|jira', 'commands': ('track')}},
    'tr-offset': {'[{h}--tr-offset <num>{e}]': {'description': 'offset, supported for action: read, apps: qc|bugzilla|jira', 'commands': ('track')}},
    'tr-page': {'[{h}--tr-page <num>{e}]': {'description': 'record page, supported for action: read, app: mantis', 'commands': ('track')}},
    'tr-per-page': {'[{h}--tr-per-page <num>{e}]': {'description': 'records per page, supported for action: read, app: mantis', 'commands': ('track')}},
    'tr-params': {'[{h}--tr-params <dict>{e}]': {'description': 'record parameters, name1:value,name2:value,... , supported for actions: create|update', 'commands': ('track')}},
    'tr-path': {'[{h}--tr-path <path>{e}]': {'description': 'directory path, dir1/dir2/... , supported for use cases: read/create folder|read/create test set|create test|read/create suite, apps: qc|testlink', 'commands': ('track')}},
    'tr-steps': {'[{h}--tr-steps <list>{e}]': {'description': 'test steps delimited by |, step parameters use dictionary form, name1:value,name2:value,...|name1:value,name2:value,... , supported for action: create, app: testlink', 'commands': ('track')}},

    # standalone with option profile trackapps
    'app': {'{h}--app qc|bugzilla|mantis|trac|testlink{e}': {'description': 'application', 'commands': ('run')}},
    'action': {'{h}--action read|create|update|delete{e}': {'description': 'action, delete supported for apps: qc|mantis|trac', 'commands': ('run')}},
    'type': {'[{h}--type defect|test-folder|test|test-set-folder|test-set|test-instance|test-suite|test-plan|build{e}]': {'description': 'entity type, default defect, supported for actions: read|create|update|delete, apps: qc|testlink', 'commands': ('run')}},
    'input': {'[{h}--input <filename>{e}]': {'description': 'filename, content is written to ticket description, supported for actions: create|update', 'commands': ('run')}},
    'output': {'[{h}--output <filename>{e}]': {'description': 'filename, writes action output, supported for action: read', 'commands': ('run')}},
    'url': {'[{h}--url <url>{e}]': {'description': 'url, configurable', 'commands': ('run')}},
    'user': {'[{h}--user <username>{e}]': {'description': 'username, configurable', 'commands': ('run')}},
    'passw': {'[{h}--passw <password>{e}]': {'description': 'password, configurable', 'commands': ('run')}},
    'dev-key': {'[{h}--dev-key <key>{e}]': {'description': 'developer key, configurable, supported for app: testlink', 'commands': ('run')}},
    'domain': {'[{h}--domain <domain>{e}]': {'description': 'domain, configurable, supported for app: qc', 'commands': ('run')}},
    'project': {'[{h}--project <project>{e}]': {'description': 'project, configurable, supported for apps: qc|mantis|trac|jira|testlink', 'commands': ('run')}},
    'id': {'[{h}--id <num>{e}]': {'description': 'record id, supported for actions: read|update|delete', 'commands': ('run')}},
    'fields': {'[{h}--fields <list>{e}]': {'description': 'requested fields, name1,name2,... , supported for action: read', 'commands': ('run')}},
    'query': {'[{h}--query <expression>{e}]': {'description': 'query, supported for action: read, apps: qc|bugzilla|trac|jira', 'commands': ('run')}},
    'order-by': {'[{h}--order-by <expression>{e}]': {'description': 'record ordering, name1:direction,name2:direction,... , direction asc|desc, supported for action: read, app: qc', 'commands': ('run')}},
    'limit': {'[{h}--limit <num>{e}]': {'description': 'limit, supported for action: read, apps: qc|bugzilla|jira', 'commands': ('run')}},
    'offset': {'[{h}--offset <num>{e}]': {'description': 'offset, supported for action: read, apps: qc|bugzilla|jira', 'commands': ('run')}},
    'page': {'[{h}--page <num>{e}]': {'description': 'record page, supported for action: read, app: mantis', 'commands': ('run')}},
    'per-page': {'[{h}--per-page <num>{e}]': {'description': 'records per page, supported for action: read, app: mantis', 'commands': ('run')}},
    'params': {'[{h}--params <dict>{e}]': {'description': 'record parameters, name1:value,name2:value,... , supported for actions: create|update', 'commands': ('run')}},
    'path': {'[{h}--path <path>{e}]': {'description': 'directory path, dir1/dir2/... , supported for use cases: read/create folder|read/create test set|create test|read/create suite, apps: qc|testlink', 'commands': ('run')}},
    'steps': {'[{h}--steps <list>{e}]': {'description': 'test steps delimited by |, step parameters use dictionary form, name1:value,name2:value,...|name1:value,name2:value,... , supported for action: create, app: testlink', 'commands': ('run')}}
}



# -*- coding: utf-8 -*-
"""Client for HP Quality Center

.. module:: trackapps.qc
   :platform: Unix
   :synopsis: Client for HP Quality Center
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
track_before_connect
track_after_connect
track_before_read
track_after_read
track_before_create
track_after_create
track_before_update
track_after_update
track_before_delete
track_after_delete
track_before_read_folder
track_after_read_folder
track_before_read_set
track_after_read_set

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from hydratk.lib.network.rest.client import RESTClient
from lxml.etree import Element, SubElement, tostring
from sys import version_info

config = {
    'sign_in': '/authentication-point/authenticate',
    'auth': '/rest/site-session',
    'sign_out': '/authentication-point/logout',
    'rest': '/rest/domains/{0}/projects/{1}/{2}s',
    'entities': ['defect', 'test-folder', 'test', 'test-set-folder', 'test-set', 'test-instance']
}


class Client(object):
    """Class Client
    """

    _mh = None
    _client = None
    _url = None
    _user = None
    _passw = None
    _domain = None
    _project = None
    _cookie = None
    _return_fields = {}
    _default_values = {}
    _is_connected = None

    def __init__(self):
        """Class constructor

        Called when the object is initialized  

        Args:
           none 

        """

        self._mh = MasterHead.get_head()
        self._client = RESTClient()

        cfg = self._mh.cfg['Extensions']['TrackApps']['qc']
        if ('return_fields' in cfg and cfg['return_fields'] != None):
            self._return_fields = cfg['return_fields']
            if (self._return_fields.__class__.__name__ == 'dict'):
                for key, value in cfg['return_fields'].items():
                    if (value != None):
                        self._return_fields[key] = value.split(',')
        if ('default_values' in cfg and cfg['default_values'] != None):
            default_values = {}
            for entity, expr in cfg['default_values'].items():
                default_values[entity] = {}
                if (expr != None):
                    params = expr.split('#')
                    for param in params:
                        conf = param.split('%')
                        default_values[entity][conf[0]] = conf[1]
            self._default_values = default_values
        if ('url' in cfg and cfg['url'] != None):
            self._url = cfg['url']
        if ('user' in cfg and cfg['user'] != None):
            self._user = cfg['user']
        if ('passw' in cfg and cfg['passw'] != None):
            self._passw = cfg['passw']
        if ('domain' in cfg and cfg['domain'] != None):
            self._domain = cfg['domain']
        if ('project' in cfg and cfg['project'] != None):
            self._project = cfg['project']

    @property
    def client(self):
        """ client property getter """

        return self._client

    @property
    def url(self):
        """ url property getter """

        return self._url

    @property
    def user(self):
        """ user property getter """

        return self._user

    @property
    def passw(self):
        """ passw property getter """

        return self._passw

    @property
    def domain(self):
        """ domain property getter """

        return self._domain

    @property
    def project(self):
        """ project property getter """

        return self._project

    @property
    def cookie(self):
        """ cookie property getter """

        return self._cookie

    @property
    def return_fields(self):
        """ return_fields property getter """

        return self._return_fields

    @property
    def default_values(self):
        """ default_values property getter """

        return self._default_values

    @property
    def is_connected(self):
        """ is_connected property getter """

        return self._is_connected

    def connect(self, url=None, user=None, passw=None, domain=None, project=None):
        """Method connects to QC

        Args:    
           url (str): URL        
           user (str): username
           passw (str): password
           domain (str): QC domain
           project (str): QC project

        Returns:
           bool: result

        Raises:
           event: track_before_connect
           event: track_after_connect

        """

        message = 'url:{0}, user:{1}, passw:{2}, domain:{3}, project:{4}'.format(
            url, user, passw, domain, project)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_connecting', message), self._mh.fromhere())

        if (url == None):
            url = self._url
        if (user == None):
            user = self._user
        if (passw == None):
            passw = self._passw
        if (domain == None):
            domain = self._domain
        if (project == None):
            project = self._project

        ev = event.Event(
            'track_before_connect', url, user, passw, domain, project)
        if (self._mh.fire_event(ev) > 0):
            url = ev.argv(0)
            user = ev.argv(1)
            passw = ev.argv(2)
            domain = ev.argv(3)
            project = ev.argv(4)

        if (ev.will_run_default()):
            self._url = url
            self._user = user
            self._passw = passw
            self._domain = domain
            self._project = project

            url = self._url + config['sign_in']
            res, body = self._client.send_request(
                url, user=self._user, passw=self._passw, method='POST')
            if (res == 200):
                url, cookie = self._url + \
                    config['auth'], self._client.get_header('set-cookie')
                headers = {'Cookie': cookie}
                res, body = self._client.send_request(
                    url, user=self._user, passw=self._passw, method='POST', headers=headers)

        result = False
        if (res in [200, 201]):

            self._cookie = '{0}, {1}'.format(
                cookie, self._client.get_header('set-cookie'))
            if (self._cookie != None):
                self._is_connected = True
                self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                    'track_connected'), self._mh.fromhere())
                ev = event.Event('track_after_connect')
                self._mh.fire_event(ev)
                result = True
            else:
                self._mh.demsg('htk_on_error', self._mh._trn.msg(
                    'track_missing_cookie'), self._mh.fromhere())
        else:
            error = body.Title if (hasattr(body, 'Title')) else body
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, error), self._mh.fromhere())

        return result

    def disconnect(self):
        """Method disconnects from QC

        Args:   
           none 

        Returns:
           bool: result

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_disconnecting'), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        url = self._url + config['sign_out']
        res, body = self._client.send_request(
            url, method='POST', headers={'Cookie': self._cookie})

        result = False
        if (res == 200):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_disconnected'), self._mh.fromhere())
            self._cookie = None
            self._is_connected = False
            result = True
        else:
            error = body.Title if (hasattr(body, 'Title')) else body
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, error), self._mh.fromhere())

        return result

    def read(self, id=None, entity='defect', fields=None, query=None, order_by=None, limit=None, offset=None):
        """Method reads records

        Args: 
           id (int): record id
           entity (str): entity type    
           fields (list): fields to be returned, default all
           query (str): record query
           order_by (dict): record ordering, key - field, value - direction asc|desc 
           limit (int): record count    
           offset (int): record offset

        Returns:
           tuple: result (bool), records (list of dictionary)

        Raises:
           event: track_before_read
           event: track_after_read

        """

        if (entity not in config['entities']):
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_unknown_entity', entity), self._mh.fromhere())
            return (False, None)

        message = 'entity:{0}, id:{1}, fields:{2}, query:{3}, order_by:{4}, limit:{5}, offset:{6}'.format(
            entity, id, fields, query, order_by, limit, offset)
        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_reading', message), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        if (fields == None and entity in self._return_fields and self._return_fields[entity] != None):
            fields = self._return_fields[entity]

        ev = event.Event(
            'track_before_read', entity, id, fields, query, order_by, limit, offset)
        if (self._mh.fire_event(ev) > 0):
            entity = ev.argv(0)
            id = ev.argv(1)
            fields = ev.argv(2)
            query = ev.argv(3)
            order_by = ev.argv(4)
            limit = ev.argv(5)
            offset = ev.argv(6)

        if (ev.will_run_default()):

            params = {}
            if (fields != None and len(fields) > 0):
                param = ""
                for field in fields:
                    param += field + ','
                params['fields'] = param[:-1]
            if (query != None):
                params['query'] = query
            if (id != None):
                params['query'] = '{ID[%s]}' % id
            if (order_by != None and len(order_by.keys()) > 0):
                param = ""
                for field, direction in order_by.items():
                    if (direction == 'asc'):
                        param += '+' + field + ','
                    elif (direction == 'desc'):
                        param += '-' + field + ','
                params['order-by'] = param[:-1]
            if (limit != None):
                params['limit'] = limit
            if (offset != None):
                params['offset'] = offset

            url = self._url + \
                config['rest'].format(self._domain, self._project, entity)
            headers = {'Cookie': self._cookie}
            res, body = self._client.send_request(url, method='GET', headers=headers, params=params,
                                                  content_type='xml')

        result = False
        records = None
        if (res == 200):
            records = []
            if (int(body.get('TotalResults')) > 0):
                for ent in body.Entity:
                    record = {}
                    for field in ent.Fields.Field:
                        key = field.get('Name')
                        value = field.Value if (
                            hasattr(field, 'Value')) else None
                        if (fields == None or key in fields):
                            record[key] = value
                    records.append(record)

            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_read', len(records)), self._mh.fromhere())
            ev = event.Event('track_after_read')
            self._mh.fire_event(ev)
            result = True
        else:
            error = body.Title if (hasattr(body, 'Title')) else body
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, error), self._mh.fromhere())

        return (result, records)

    def create(self, entity='defect', params={}):
        """Method creates record

        Args: 
           entity (str): entity type
           params (dict): record content, key - field name, value - field value

        Returns:
           int: record id

        Raises:
           event: track_before_create
           event: track_after_create

        """

        if (entity not in config['entities']):
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_unknown_entity', entity), self._mh.fromhere())
            return None

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_creating', entity, params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        if (entity in self._default_values and self._default_values[entity] != None):
            for key, value in self._default_values[entity].items():
                if (key not in params):
                    params[key] = value

        ev = event.Event('track_before_create', entity, params)
        if (self._mh.fire_event(ev) > 0):
            entity = ev.argv(0)
            params = ev.argv(1)

        if (ev.will_run_default()):

            root = Element('Entity')
            root.set('Type', entity)
            el_fields = SubElement(root, 'Fields')
            for key, value in params.items():
                elem = SubElement(el_fields, 'Field')
                elem.set('Name', key)
                SubElement(elem, 'Value').text = str(value).decode(
                    'utf8') if (version_info[0] == 2) else str(value)
            body = tostring(root)

            url = self._url + \
                config['rest'].format(self._domain, self._project, entity)
            headers = {'Cookie': self._cookie}
            res, body = self._client.send_request(url, method='POST', headers=headers, body=body,
                                                  content_type='xml')
        id = None
        if (res in (200, 201)):
            id = int(
                body.xpath("//Entity/Fields/Field[@Name='id']/Value/text()")[0])
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_created', id), self._mh.fromhere())
            ev = event.Event('track_after_create')
            self._mh.fire_event(ev)
        else:
            error = body.Title if (hasattr(body, 'Title')) else body
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, error), self._mh.fromhere())

        return id

    def update(self, id, entity='defect', params={}):
        """Method updates record

        Args: 
           id (int): entity id
           entity (str): entity type         
           params (dict): record content, key - field name, value - field value 

        Returns:
           bool: result

        Raises:
           event: track_before_update
           event: track_after_update

        """

        if (entity not in config['entities']):
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_unknown_entity', entity), self._mh.fromhere())
            return False

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_updating', entity, id, params), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_update', entity, id, params)
        if (self._mh.fire_event(ev) > 0):
            entity = ev.argv(0)
            id = ev.argv(1)
            params = ev.argv(2)

        if (ev.will_run_default()):

            root = Element('Entity')
            root.set('Type', entity)
            el_fields = SubElement(root, 'Fields')
            for key, value in params.items():
                elem = SubElement(el_fields, 'Field')
                elem.set('Name', key)
                SubElement(elem, 'Value').text = str(value).decode(
                    'utf8') if (version_info[0] == 2) else str(value)
            body = tostring(root)

            url = self._url + \
                config['rest'].format(
                    self._domain, self._project, entity) + '/' + str(id)
            headers = {'Cookie': self._cookie}
            res, body = self._client.send_request(url, method='PUT', headers=headers, body=body,
                                                  content_type='xml')
        result = False
        if (res == 200):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_updated', id), self._mh.fromhere())
            ev = event.Event('track_after_update')
            self._mh.fire_event(ev)
            result = True
        else:
            error = body.Title if (hasattr(body, 'Title')) else body
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, error), self._mh.fromhere())

        return result

    def delete(self, id, entity='defect'):
        """Method deletes record

        Args: 
           id (int): entity id
           entity (str): entity type          

        Returns:
           bool: result

        Raises:
           event: track_before_delete
           event: track_after_delete

        """

        if (entity not in config['entities']):
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_unknown_entity', entity), self._mh.fromhere())
            return False

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
            'track_deleting', entity, id), self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False

        ev = event.Event('track_before_delete', entity, id)
        if (self._mh.fire_event(ev) > 0):
            entity = ev.argv(0)
            id = ev.argv(1)

        if (ev.will_run_default()):

            url = self._url + \
                config['rest'].format(
                    self._domain, self._project, entity) + '/' + str(id)
            headers = {'Cookie': self._cookie}
            res, body = self._client.send_request(
                url, method='DELETE', headers=headers, content_type='xml')

        result = False
        if (res == 200):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_deleted', id), self._mh.fromhere())
            ev = event.Event('track_after_delete')
            self._mh.fire_event(ev)
            result = True
        else:
            error = body.Title if (hasattr(body, 'Title')) else body
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_error', res, error), self._mh.fromhere())

        return result

    def read_test_folder(self, path, entity='test-folder'):
        """Method reads tests under test folder

        Args: 
           path (str): folder path   
           entity (str): entity type, test-folder|test-set-folder

        Returns:
           tuple: result (bool), records (dict), key - test folder, value - list of tests

        Raises:
           event: track_before_read_folder
           event: track_after_read_folder

        """

        if (entity not in ['test-folder', 'test-set-folder']):
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_unknown_entity', entity), self._mh.fromhere())
            return (False, None)

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_reading_folder', path, entity),
                      self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        ev = event.Event('track_before_read_folder', path, entity)
        if (self._mh.fire_event(ev) > 0):
            path = ev.argv(0)
            entity = ev.argv(1)

        if (ev.will_run_default()):

            folder = self._get_folder(path, entity)
            if (folder == None):
                return (False, None)

            hier_path = folder['hierarchical-path']
            if (hier_path != None):
                fields = ['id', 'name', 'parent-id']
                query = '{hierarchical-path[%s*]}' % hier_path
                res, folders = self.read(
                    entity=entity, fields=fields, query=query)

                top_id = folder['id']
                names_new = []
                for i in range(0, len(folders)):

                    if (folders[i]['id'] == top_id):
                        names_new.append(path)
                    else:
                        parent_id = folders[i]['parent-id']
                        name = folders[i]['name']
                        while (parent_id != top_id):

                            j = 0
                            while (folders[j]['id'] != parent_id):
                                j += 1
                            parent_id = folders[j]['parent-id']
                            name = '{0}/{1}'.format(folders[j]['name'], name)

                        names_new.append('{0}/{1}'.format(path, name))

                for i in range(0, len(folders)):
                    folders[i]['name'] = names_new[i]

        if (res):
            tests = {}
            cnt = 0
            entity = 'test' if (entity == 'test-folder') else 'test-set'
            fields = None if (entity == 'test-folder') else ['id', 'name']
            for rec in folders:
                query = '{parent-id[%d]}' % rec['id']
                res, records = self.read(
                    entity=entity, fields=fields, query=query)
                if (res):
                    tests[rec['name']] = records
                    cnt += len(records)

            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_folder_read', cnt), self._mh.fromhere())
            ev = event.Event('track_after_read_folder')
            self._mh.fire_event(ev)

            return (True, tests)

        else:
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_missing_hier', folder['name']), self._mh.fromhere())
            return (False, None)

    def create_test_folder(self, path, name, entity='test-folder'):
        """Method creates test folder on path

        Args: 
           path (str): folder path
           name (str): folder name
           entity (str): entity type, test-folder|test-set-folder

        Returns:
           int: folder id

        """

        if (entity not in ['test-folder', 'test-set-folder']):
            self._mh.demsg('htk_on_error', self._mh._trn.msg(
                'track_unknown_entity', entity), self._mh.fromhere())
            return None

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        folder = self._get_folder(path, entity)
        if (folder == None):
            return None

        params = {'name': name, 'parent-id': folder['id']}
        return self.create(entity, params)

    def read_test_set(self, id):
        """Method reads tests under test set

        Args: 
           id (int): test set id

        Returns:
           tuple: result (bool), tests (list of dict)

        Raises:
           track_before_read_set
           track_after_read_set

        """

        self._mh.demsg('htk_on_debug_info', self._mh._trn.msg('track_reading_set', id),
                      self._mh.fromhere())

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return False, None

        ev = event.Event('track_before_read_set', id)
        if (self._mh.fire_event(ev) > 0):
            id = ev.argv(0)

        if (ev.will_run_default()):
            fields = ['test-id', 'status', 'exec-date', 'actual-tester']
            query = '{cycle-id[%d]}' % id
            res, tests = self.read(
                entity='test-instance', fields=fields, query=query)

        if (res):
            self._mh.demsg('htk_on_debug_info', self._mh._trn.msg(
                'track_set_read', len(tests)), self._mh.fromhere())
            ev = event.Event('track_after_read_set')
            self._mh.fire_event(ev)

            return (True, tests)
        else:
            return (False, None)

    def create_test_set(self, path, params={}):
        """Method creates test set on path

        Args: 
           path (str): folder path
           params (dict): test content, key - field name, value - field value  

        Returns:
           int: test set id

        """

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        folder = self._get_folder(path, 'test-set-folder')
        if (folder == None):
            return None

        params['parent-id'] = folder['id']
        return self.create('test-set', params)

    def create_test(self, path, params={}):
        """Method creates test on path

        Args: 
           path (str): folder path
           params (dict): test content, key - field name, value - field value    

        Returns:
           int: test id

        """

        if (not self._is_connected):
            self._mh.demsg('htk_on_warning', self._mh._trn.msg(
                'track_not_connected'), self._mh.fromhere())
            return None

        folder = self._get_folder(path)
        if (folder == None):
            return None

        params['parent-id'] = folder['id']
        return self.create('test', params)

    def _get_folder(self, path, entity='test-folder'):
        """Method gets folder from hierarchical path

        Args: 
           path (str): folder path, Folder1/Folder2/...    
           entity (str): entity type, test-folder|test-set-folder

        Returns:
           dict

        """

        folders = path.split('/')
        fields = ['id', 'name', 'hierarchical-path']
        parent_id = 0

        if (folders[0] == 'Root'):
            folders = folders[1:]

        cnt = len(folders)
        for i in range(0, cnt):
            query = '{parent-id[%d]}' % parent_id
            res, records = self.read(entity=entity, fields=fields, query=query)

            if (res):
                found = False
                j = 0
                while (not found and j < len(records)):
                    record = records[j]
                    if (record['name'] == folders[i]):
                        parent_id = record['id']
                        found = True
                    j += 1

            if (not found):
                self._mh.demsg('htk_on_error', self._mh._trn.msg('track_unknown_folder', folders[i]),
                              self._mh.fromhere())
                return None

        return record
